#!/usr/bin/env python3.2
#
# Copyright (c) Net24 Limited, Christchurch, New Zealand 2011-2012
#       and     Voyager Internet Ltd, New Zealand, 2012-2013
#
#    This file is part of py-magcode-core.
#
#    Py-magcode-core is free software: you can redistribute it and/or modify
#    it under the terms of the GNU  General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Py-magcode-core is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU  General Public License for more details.
#
#    You should have received a copy of the GNU  General Public License
#    along with py-magcode-core.  If not, see <http://www.gnu.org/licenses/>.
#
"""Main DNS Management Daemon Code

Program implemented by subclassing magcode.core.process.Process, and 
replacing the main() method.
"""

import os
import os.path
import errno
import sys
import pwd
import time
import copy
import signal
import gc
import json

import psutil

from magcode.core.process import ProcessDaemon
from magcode.core.process import SignalHandler
from magcode.core.globals_ import *
from magcode.core.database import *
from magcode.core.database.event import EventQueue
from magcode.core.utility import get_numeric_setting
from magcode.core.utility import get_boolean_setting
# import to pull in and init ProcessSM
import magcode.core.database.process_sm
# import to pull in and init ZoneSMs
import dms.database.zone_sm
from magcode.core.utility import connect_test_address
from dms.database.master_sm import recalc_machine_dns_server_info
from dms.database.server_sm import init_soaquery_rcodes
# import to fully init settings for config file DEFAULT section
from dms.globals_ import update_engine
from dms.dyndns_update import DynDNSUpdate
from dms.exceptions import DynDNSCantReadKeyError

USAGE_MESSAGE = "Usage: %s [-dhv] [-c config_file]"
COMMAND_DESCRIPTION = "DMS DNS Management Daemon"


class SIGUSR1Handler(SignalHandler):
    """
    Handle a SIGUSR1 signal.

    Just make action() return False to wake loop
    """
    def action(self):
        log_info('SIGUSR1 received - running event queue.')
        return False

class DmsDMDProcess(ProcessDaemon):
    """
    Process Main Daemon class
    """
    def __init__(self, *args, **kwargs):
        super().__init__(usage_message=USAGE_MESSAGE,
            command_description=COMMAND_DESCRIPTION, *args, **kwargs)

    def init_signals(self):
        """
        Initialise signal handlers for the daemon
        """
        super().init_signals()
        self.register_signal_handler(signal.SIGUSR1, SIGUSR1Handler())
    
    def init_master_dns_address(self):
        """
        Master dns server setting in an IP addr

        Results determined by getaddrinfo(3) and thus by /etc/hosts contents, 
        or else DNS if hostname not in /etc/hosts!
        """
        test_hostname = settings['master_dns_server']
        if not test_hostname:
            test_hostname = socket.getfqdn()
        connect_retry_wait = get_numeric_setting('connect_retry_wait', float)
        exc_msg = ''
        for t in range(3):
            try:
                # Transform any hostname to an IP address
                settings['master_dns_server'] = connect_test_address(
                                                test_hostname,
                                                port=settings['master_dns_port'])
                break
            except(IOError, OSError) as exc:
                exc_msg = str(exc)
                time.sleep(connect_retry_wait)
                continue
        else:
            log_error("Testing master DNS server IP address '%s:%s' - %s" 
                        % (test_hostname, settings['master_dns_port'], exc_msg))
            systemd_exit(os.EX_NOHOST, SDEX_CONFIG)
        # If we get here without raising an exception, we can talk to
        # the server address (mostly)
        return

    def init_master_dns_server_data(self):
        """
        Read in configuration values for these, and then process them

        This is a bit messy, but it does the job just here.
        """
        # We use config file initially to set list
        this_servers_addresses = settings['this_servers_addresses']
        if isinstance(this_servers_addresses, str):
            try:
                this_servers_addresses = settings['this_servers_addresses']\
                                        .replace(',', ' ')\
                                        .replace("'", ' ')\
                                        .replace('"', ' ')\
                                        .replace('[', ' ')\
                                        .replace(']', ' ')\
                                        .strip().split()
            except ValueError as exc:
                log_error("Could not parse 'this_servers_addresses' to obtain"
                        " list of this servers DNS listening addresses - %s"
                            % str(exc))
                systemd_exit(os.EX_CONFIG, SDEX_CONFIG)
        settings['this_servers_addresses'] = this_servers_addresses
        # Recalculate host information - this will do nothing 
        # if 'ifconfig -a' et al won't work
        ifconfig_exc = (True if not settings['this_servers_addresses'] 
                            else False)
        try:
            db_session = sql_data['scoped_session_class']()
            recalc_machine_dns_server_info(db_session, ifconfig_exc)
            db_session.commit()
        except Exception as exc:
            log_error(str(exc))
            systemd_exit(os.EX_UNAVAILABLE, SDEX_NOTRUNNING)
        log_info("List of local IPs, 'this_servers_addresses' - %s"
                    % ', '.join(settings['this_servers_addresses']))
        log_info("Master DNS server on this machine, 'master_dns_server' - %s"
                    % settings['master_dns_server'])

    def init_update_engines(self):
        """
        Initialise the update engines used
        """
        connect_retry_wait = get_numeric_setting('connect_retry_wait', float)
        error_str = ''
        for t in range(3):
            try:
                dyndns_engine = DynDNSUpdate(settings['dns_server'],
                                        settings['dyndns_key_file'],
                                        settings['dyndns_key_name'])
                break
            except (DynDNSCantReadKeyError, IOError, OSError) as exc:
                error_str = ("Can't connect to named for dynamic updates - %s" 
                                % str(exc))
                time.sleep(connect_retry_wait)
                continue
        # Process above error...
        else:
            log_error("%s" % error_str)
            systemd_exit(os.EX_NOHOST, SDEX_CONFIG)
        update_engine['dyndns'] = dyndns_engine

    def do_garbage_collect(self):
        """
        Do Resource Release exercise at low memory threshold, blow up over max
        """
        error_str = ''
        try:
            rss_mem_usage = (float(self.proc_monitor.memory_info().rss)
                                    /1024/1024)
        except AttributeError:
            # Deal with a change in name of get_memory_info() method
            rss_mem_usage = (float(self.proc_monitor.get_memory_info().rss)
                                    /1024/1024)
        except Exception as exc:
            error_str = str(exc)
        # Process above error...
        if (error_str):
            log_error("Error obtaining resource usage - %s" % error_str) 
            systemd_exit(os.EX_SOFTWARE, SDEX_NOTRUNNING)
        memory_exec_threshold = get_numeric_setting('memory_exec_threshold', float)
        if (rss_mem_usage > memory_exec_threshold):
            log_warning('Memory exec threshold %s MB reached, actual %s MB - execve() to reclaim.'
                        % (memory_exec_threshold, rss_mem_usage))
            file_path = os.path.join(sys.path[0], sys.argv[0])
            file_path = os.path.normpath(file_path)
            os.execve(file_path, sys.argv, os.environ)
        else:
            # Spend idle time being RAM thrifty...
            gc.collect()
            return

    def main_process(self):
        """Main process for dmsdmd
        """
        
        if (settings['rpdb2_wait']):
            # a wait to attach with rpdb2...
            log_info('Waiting for rpdb2 to attach.')
            time.sleep(float(settings['rpdb2_wait']))

        log_info('program starting.')
        log_debug("The daemon_canary is: '%s'" % settings['daemon_canary'])
        # Do a nice output message to the log
        pwnam = pwd.getpwnam(settings['run_as_user'])
        log_debug("PID: %s daemon: '%s' User: '%s' UID: %d GID %d" 
                % (os.getpid(), self.i_am_daemon(), pwnam.pw_name,
                    os.getuid(), os.getgid()))

        # Check we can reach DNS server
        self.init_update_engines()
        
        # Initialise ServerSM rcodes from settings
        init_soaquery_rcodes()

        # Initialize master dns address if required
        self.init_master_dns_address()

        # Connect to database, intialise SQL Alchemy
        setup_sqlalchemy()

        # Initialize master DNS server data
        self.init_master_dns_server_data()

        # Create a queue
        event_queue = EventQueue()

        # Create a Process object so that we can check in on ourself resource
        # wise
        self.proc_monitor = psutil.Process(pid=os.getpid())

        # Initialise  a few nice things for the loop
        debug_mark = get_boolean_setting('debug_mark') 
        sleep_time = get_numeric_setting('sleep_time', float)
        # test Read this value...
        master_hold_timeout = get_numeric_setting('master_hold_timeout', float)

        if (settings['memory_debug']):
            # Turn on memory debugging
            log_info('Turning on GC memory debugging.')
            gc.set_debug(gc.DEBUG_LEAK)

        # Process Main Loop
        while (self.check_signals()):
            
            event_queue.process_queue()
            if event_queue.queue_empty():
                self.do_garbage_collect()
            if debug_mark:
                log_debug("----MARK---- sleep(%s) seconds ----"
                        % sleep_time) 
            time.sleep(sleep_time)

        log_info('Exited main loop - process terminating normally.')
        sys.exit(os.EX_OK)


if (__name__ is "__main__"):
    exit_code = DmsDMDProcess(sys.argv, len(sys.argv))
    sys.exit(exit_code)



#!/usr/bin/env python3.2
#
# Copyright (c) Net24 Limited, Christchurch, New Zealand 2011-2012
#       and     Voyager Internet Ltd, New Zealand, 2012-2013
#
#    This file is part of py-magcode-core.
#
#    Py-magcode-core is free software: you can redistribute it and/or modify
#    it under the terms of the GNU  General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Py-magcode-core is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU  General Public License for more details.
#
#    You should have received a copy of the GNU  General Public License
#    along with py-magcode-core.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Module for ZoneEngine base class
"""


import datetime
import re
import socket
from io import StringIO

from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.orm.exc import MultipleResultsFound
from sqlalchemy import desc
from sqlalchemy import text
from pyparsing import ParseBaseException

from magcode.core.globals_ import *
from dms.globals_ import *
from magcode.core.database import *
from magcode.core.database import *
from magcode.core.database.event import find_events
from magcode.core.database.event import ESTATE_FAILURE
from magcode.core.database.event import create_event
from magcode.core.database.event import Event
# import all possible types used here so that they intialise in zone-tool
from dms.database.master_sm import zone_sm_dnssec_schedule
from dms.database.zone_sm import ZoneSM
from dms.database.zone_sm import ZoneSMEdit
from dms.database.zone_sm import ZoneSMEditExit
from dms.database.zone_sm import ZoneSMEditTimeout
from dms.database.zone_sm import ZoneSMEditLockTickle
from dms.database.zone_sm import ZoneSMEditUpdate
from dms.database.zone_sm import ZoneSMUpdate
from dms.database.zone_sm import ZoneSMEditSaved
from dms.database.zone_sm import ZoneSMEnable
from dms.database.zone_sm import ZoneSMDisable
from dms.database.zone_sm import ZoneSMDoReconfig
from dms.database.zone_sm import ZoneSMDoBatchConfig
from dms.database.zone_sm import ZoneSMDoConfig
from dms.database.zone_sm import ZoneSMEditSavedNoLock
from dms.database.zone_sm import ZoneSMDelete
from dms.database.zone_sm import ZoneSMUndelete
from dms.database.zone_sm import ZoneSMDoReset
from dms.database.zone_sm import ZoneSMDoRefresh
from dms.database.zone_sm import ZoneSMDoDestroy
from dms.database.zone_sm import ZoneSMDoSgSwap
from dms.database.zone_sm import ZoneSMDoSetAltSg
from dms.database.zone_sm import ZLSTATE_EDIT_LOCK
from dms.database.zone_sm import ZSTATE_DISABLED
from dms.database.zone_sm import ZSTATE_UNCONFIG
from dms.database.zone_sm import ZSTATE_DELETED
from dms.database.zone_sm import ZSTATE_PUBLISHED
from dms.database.zone_sm import exec_zonesm
from dms.database.zone_sm import new_zone
from dms.database.zone_sm import DynDNSZoneSM
from dms.database.zone_instance import ZoneInstance
from dms.database.zone_instance import new_zone_zi
from dms.database.resource_record import ResourceRecord
from dms.database.resource_record import data_to_rr
import dms.database.zone_cfg as zone_cfg
from dms.database.reference import Reference
from dms.database.rr_comment import RRComment
from dms.database.zone_sectag import ZoneSecTag
from dms.database.zone_sectag import list_all_sectags
# import securitytags so that sql_data is initialised
import dms.database.zone_sectag
from dms.database.master_sm import show_master_sm
from dms.database.master_sm import get_mastersm_replica_sg
from dms.database.sg_utility import list_all_sgs
from dms.database.sg_utility import find_sg_byname
from dms.database.reference import new_reference
from dms.database.reference import del_reference
from dms.database.reference import find_reference
from dms.database.reference import rename_reference
from dms.database.server_sm import ServerSM
from dms.database.server_group import ServerGroup
from magcode.core.wsgi.jsonrpc_server import InvalidParamsJsonRpcError
from dms.exceptions import *
from dms.database.zone_query import rr_query_db_raw
from dms.zone_data_util import ZoneDataUtil
from dms.dns import is_inet_domain
from dms.dns import is_network_address
from dms.dns import wellformed_cidr_network
from dms.dns import zone_name_from_network
from dms.dns import new_soa_serial_no
from dms.database.reverse_network import new_reverse_network
from dms.database.reverse_network import ReverseNetwork
from dms.zone_text_util import data_to_bind
from dms.zone_text_util import bind_to_data


class ZoneEngine(ZoneDataUtil):
    """
    Base Zone Editing/control Engine container class

    Contains common code and stub methods.
    """
    def __init__(self, time_format=None, sectag_label=None):
        """
        Initialise engine. Get a scoped DB session.
        """
        self.time_format = time_format
        self.sectag = ZoneSecTag(sectag_label=sectag_label)
        self.refresh_db_session()
        if self.sectag not in list_all_sectags(self.db_session):
            raise ZoneSecTagConfigError(self.sectag.sectag)

    def refresh_db_session(self):
        self.db_session = sql_data['scoped_session_class']()

    def rollback(self):
        self.db_session.rollback()

    def _finish_op(self):
        self.db_session.commit()
    
    def _begin_op(self):
        # Refresh SA session
        self.refresh_db_session()
        self.db_session.commit()

    _login_id_char_re = re.compile(r'^[\-_a-zA-Z0-9.@]+$')
    _login_id_start_re = re.compile(r'^[0-9a-zA-Z][\-_a-zA-Z0-9.@]*$')
    def _make_change_by(self, login_id):
        """
        Create a change_by string from a login_id
        """
        # Check that the supplied login_id is acceptable
        if not login_id:
            raise LoginIdInvalidError("a login_id must be given" )
        if not isinstance(login_id, str):
            raise LoginIdInvalidError("login_id must be a string" )
        if len(login_id) > 512:
            error_msg = "too long, must be <= 512."
            raise LoginIdInvalidError(error_msg)
        if not self._login_id_char_re.match(login_id):
            error_msg = "can only contain characters '-_a-zA-Z0-9.@'"
            raise LoginIdFormatError(login_id, error_msg)
        if not self._login_id_start_re.match(login_id):
            error_msg = "must start with 'a-zA-Z0-9'"
            raise LoginIdFormatError(login_id, error_msg)

        return login_id + '/' + self.sectag.sectag      

    def _list_zone(self, names=None, reference=None, sg_name=None,
            include_deleted=False, toggle_deleted=False,
            include_disabled=True):
        """
        Backend search for the given names.  Multiple names may be given.
        Wildcards can be used for partial matches.  No name will list all 
        zones.
        """

        def build_query(query):
            """
            Common query code
            """
            if reference:
                query = query.join(Reference)\
                        .filter(Reference.reference.ilike(reference))
            if sg_name and self.sectag.sectag == settings['admin_sectag']:
                if sg_name not in list_all_sgs(self.db_session):
                    raise NoSgFound(sg_name)
                query = query.join(ServerGroup,
                         ServerGroup.id_ == ZoneSM.sg_id)\
                        .filter(ServerGroup.name == sg_name)
            if include_deleted:
                pass
            elif toggle_deleted:
                query = query.filter(ZoneSM.state == ZSTATE_DELETED)
            else:
                query = query.filter(ZoneSM.state != ZSTATE_DELETED)
            if not include_disabled:
                query = query.filter(ZoneSM.state != ZSTATE_DISABLED)
            return(query)
       
        # Deal with SA auto-BEGIN - want fresh transaction to see fresh data
        self._begin_op()
        db_query_slice = get_numeric_setting('db_query_slice', int)
        if not names:
            # No arguments
            query = self.db_session.query(ZoneSM)
            query = build_query(query)
            query = query.order_by(ZoneSM.name)
            # Set up query so that server side curses are used, and the whole
            # zone database is not grabbed all at once, stopping extreme
            # allocation of memory...
            query = query.yield_per(db_query_slice)
            zones = []
            for z in query:
                if (self.sectag.sectag != settings['admin_sectag']
                        and self.sectag not in z.sectags):
                    continue
                zones.append(z.to_engine_brief())
            self._finish_op()
            if not zones:
                raise NoZonesFound('')
            return zones
       
        # We were given some arguments
        zones = []
        # We keep domains and labels in database lowercase
        name_pattern = ' '.join(names).lower()
        names = [x.lower() for x in names]
        names = [x.replace('*', '%') for x in names]
        names = [x.replace('?', '_') for x in names]
        # Perform limit checks to prevent RAM hoggery DOS Death By SQL
        if ('%' in names and len(names) > 1):
            raise OnlyOneLoneWildcardValid(name_pattern)
        # Check that reference is given for non admin sectag accesses
        if (self.sectag.sectag != settings['admin_sectag'] 
                and not reference):
            raise ReferenceMustBeGiven(name_pattern)
        for name in names:
            network_address_flag = is_network_address(name)
            network = wellformed_cidr_network(name, filter_mask_size=False)
            query = self.db_session.query(ZoneSM)
            if network_address_flag:
                query = query.join(ReverseNetwork)\
                        .filter( text(':name <<= reverse_networks.network'))\
                        .params(name=name)\
                        .order_by(ReverseNetwork.network)
            elif network:
                query = query.join(ReverseNetwork)\
                        .filter( text(':network >>= reverse_networks.network'))\
                        .params(network=network)\
                        .order_by(ReverseNetwork.network)
            else:
                if not name.endswith('.') and not name.endswith('%'):
                    name += '.'
                query = query.filter(ZoneSM.name.like(name))
            query = build_query(query)
            query = query.yield_per(db_query_slice)
            for z in query:
                if (self.sectag.sectag != settings['admin_sectag']
                        and self.sectag not in z.sectags):
                    continue
                zones.append(z.to_engine_brief())
        zones = sorted(zones, key=lambda zone: zone['name'])
        if not zones:
            raise NoZonesFound(name_pattern)
        self._finish_op()
        return zones

    def list_zone(self, names=None, reference=None):
        """
        1st domains level list_zone call
        """
        return self._list_zone(names=names, reference=reference)

    def list_zone_admin(self, names=None, reference=None, 
            sg_name=None, include_deleted=False, toggle_deleted=False,
            include_disabled=True):
        """
        Admin privilege list_zone()
        """
        return self._list_zone(names=names, reference=reference,
                sg_name=sg_name, include_deleted=include_deleted,
                toggle_deleted=toggle_deleted,
                include_disabled=include_disabled)

    def _get_zone_sm(self, name, zone_id=None, check_sectag=True,
            toggle_deleted=False, include_deleted=False, exact_network=False):
        """
        Get zone_sm.
        """
        db_session = self.db_session
        # We keep domains and labels in database lowercase
        name = name.lower()
        multiple_results = False
        network_address_flag = is_network_address(name)
        # Don't reassign name so that error messages follow what user supplied
        # as input
        network = wellformed_cidr_network(name)
        try:
            query = db_session.query(ZoneSM)
            if network_address_flag and not exact_network:
                query = query.join(ReverseNetwork)\
                        .filter( text(':inet <<= reverse_networks.network'))\
                        .params(inet=name)\
                        .order_by(ReverseNetwork.network.desc())
            elif network_address_flag and exact_network:
                raise ZoneNotFound(name)
            elif network and not exact_network:
                query = query.join(ReverseNetwork)\
                        .filter( text(':inet <<= reverse_networks.network'))\
                        .params(inet=network)\
                        .order_by(ReverseNetwork.network.desc())
            elif network and exact_network:
                query = query.join(ReverseNetwork)\
                        .filter( text(':inet = reverse_networks.network'))\
                        .params(inet=network)\
                        .order_by(ReverseNetwork.network.desc())
            else:
                query = query.filter(ZoneSM.name == name)
            if zone_id:
                query = query.filter(ZoneSM.id_ == zone_id)
            if include_deleted:
                pass
            elif toggle_deleted:
                query = query.filter(ZoneSM.state == ZSTATE_DELETED)
            else:
                query = query.filter(ZoneSM.state != ZSTATE_DELETED)
            if network or network_address_flag:
                query = query.limit(1)
            zone_sm = query.one()
        except NoResultFound:
            zone_sm = None
        except MultipleResultsFound:
            multiple_results = True
        # Decoupled exception traces
        if multiple_results:
            raise ZoneMultipleResults(name)
        if not zone_sm:
            raise ZoneNotFound(name)
        if not check_sectag:
            return zone_sm
        # Check security tag
        if self.sectag.sectag == settings['admin_sectag']:
            return zone_sm
        if self.sectag not in zone_sm.sectags:
            raise ZoneNotFound(name)
        return zone_sm

    def _get_zone_sm_byid(self, zone_id):
        """
        Get zone_sm.
        """
        db_session = self.db_session
        # Get active zi_id
        try:
            zone_sm = db_session.query(ZoneSM)\
                        .filter(ZoneSM.id_ == zone_id).one()
        except NoResultFound:
            zone_sm = None
        # Decoupled exception traces
        if not zone_sm:
            raise ZoneNotFoundByZoneId(zone_id)
        # Check security tag
        if self.sectag.sectag == settings['admin_sectag']:
            return zone_sm
        if self.sectag not in zone_sm.sectags:
            raise ZoneNotFoundByZoneId(zone_id)
        return zone_sm

    def _get_zi(self, zi_id):
        """
        Get zi.
        """
        db_session = self.db_session
        # Get active zi_id
        zi = self._resolv_zi_id(None, zi_id,
                        specific_zi_id=True)
        if not zi:
            raise ZiNotFound('*', zi_id)
        return zi
  
    # Parsing regexps for zi_id.  Also see _zi_id_human_str in 
    # dms.exceptions
    _zi_am_pm_str = r'am|pm|AM|PM|aM|pM|Pm|Am|a|A|p|P'
    
    _zi_adj_re = re.compile(r'^\^(-+|\++|-\S+|\+\S+)$')
    _zi_adj_minus_re = re.compile(r'^-+$')
    _zi_adj_plus_re = re.compile(r'^\++$')
    _zi_adj_minusn_re = re.compile(r'^-(\S+)$')
    _zi_adj_plusn_re = re.compile(r'^\+(\S+)$')
    
    _zi_unit_re = re.compile(r'^\@(\S+)([smhdw])$')
    _zi_ddmmyyyy_hhmm_re = re.compile(r'^(\S+)\/(\S+)\/(\S+),(\S+):(\S+?)('
                                + _zi_am_pm_str + r'){0,1}$')
    _zi_iso_date_hhmm_re = re.compile(r'^(\S+)-(\S+)-(\S+),(\S+):(\S+?)(' 
                                            + _zi_am_pm_str + r'){0,1}$')
    _zi_ddmmyyyy_re = re.compile(r'^(\S+)\/(\S+)\/(\S+)$')
    _zi_iso_date_re = re.compile(r'^(\S+)-(\S+)-(\S+)$')
    _zi_ddslashmm_re = re.compile(r'^(\S+)\/(\S+)$')
    _zi_hhmm_re = re.compile(r'^(\S+):(\S+?)(' + _zi_am_pm_str + r'){0,1}$')
    _zi_int_adj_re = re.compile(r'^(\S+)(-+|\++|-\S+|\+\S+)$')
    def _resolv_zi_id(self, zone_sm, zi_id, specific_zi_id=False):
        """
        Resolve a zi_id from a string form
        """
        def new_query():
            if not zone_sm:
                query = db_session.query(ZoneInstance)
            else:
                query = zone_sm.all_zis
            query = query.yield_per(db_query_slice)
            return query

        def resolv_adj_str(adj_str):
            nonlocal query
            minusn_match = self._zi_adj_minusn_re.search(adj_str)
            plusn_match = self._zi_adj_plusn_re.search(adj_str)
            try:
                if self._zi_adj_minus_re.search(adj_str):
                    delta = -1 * len(adj_str)
                elif self._zi_adj_plus_re.search(adj_str):
                    delta = len(adj_str)
                elif minusn_match:
                    delta = -1 * int(minusn_match.group(1))
                elif plusn_match:
                    delta = int(plusn_match.group(1))
                else:
                    raise ZiIdAdjStringSyntaxError(zi_id)
            except ValueError:
                raise ZiIdAdjStringSyntaxError(zi_id)
            
            # A bit of SQL magic to get offset from pivot ID
            subq = db_session.query(ZoneInstance)\
                    .filter(ZoneInstance.id_ == pivot_zi_id).subquery()
            if delta < 0:
                query = query.filter(ZoneInstance.ctime <= subq.c.ctime)\
                        .order_by(ZoneInstance.ctime.desc())
                delta *= -1
            else:
                query = query.filter(ZoneInstance.ctime >= subq.c.ctime)\
                        .order_by(ZoneInstance.ctime.asc())
            try:
                result = query.offset(delta).limit(1).one()
            except NoResultFound:
                result = None
            return result

        def ctime_query(target_ctime):
            nonlocal query
            query = query.filter(ZoneInstance.ctime <= target_ctime)\
                        .order_by(ZoneInstance.ctime.desc()).limit(1)
            try:
                result = query.one()
            except NoResultFound:
                result = None
            return result

        def do_year_date_time(regexp_match, date_exception, iso_format_date):
            """
            Work out target_ctime, given a complete date 
            """
            match_args = regexp_match.groups()
            try:
                if iso_format_date:
                    year = int(match_args[0])
                    month = int(match_args[1])
                    day = int(match_args[2])
                else:
                    day = int(match_args[0])
                    month = int(match_args[1])
                    year = int(match_args[2])
            except ValueError:
                raise date_exception(zi_id)

            if len(match_args) == 3:
                # time not given, assume midnight
                hour = 0
                minute = 0
            else:
                try:
                    hour = int(match_args[3])
                    minute = int(match_args[4])
                except ValueError:
                    raise ZiIdHhMmSyntaxError(zi_id)
                # Process AM/PM
                if len(match_args) > 5 and match_args[5]:
                    am_pm = match_args[5].lower()
                    if (am_pm.startswith('p') and hour < 12):
                        hour += 12

            # Sort out 2 digit years
            if (70 <= year <= 99):
                year += 1900
            elif ( 0 <= year < 70):
                year += 2000
            # Use DB server as time base
            now = db_clock_time(db_session)
            try:
                target_time = datetime.time(hour, minute, tzinfo=now.tzinfo)
                target_date = datetime.date(year, month, day)
                target_ctime = datetime.datetime.combine(target_date,
                        target_time)
            except ValueError:
                raise ZiIdHhMmSyntaxError(zi_id)
            return ctime_query(target_ctime)

        # Easy as pie and basket case, quicker to do first
        if zone_sm:
            if not zi_id or zi_id == '^':
                return zone_sm.zi
        else:
            if not zi_id:
                return None
        
        db_session = self.db_session
        db_query_slice = get_numeric_setting('db_query_slice', int)
        # Fast path - check and see if zi_id is a straight integer
        query = new_query()
        try:
            zi_id = int(zi_id)
            zi = query.filter(ZoneInstance.id_ == zi_id).one()
            return zi
        except NoResultFound:
            return None
        except ValueError:
            pass
        if specific_zi_id:
            return None
        
        # Put the brakes on
        # Only zone_tool related parsing from here on 
        if (self.sectag.sectag != settings['admin_sectag'] 
                and settings['process_name'] != 'zone_tool'):
            return None

        # Try
        match_adj = self._zi_adj_re.search(zi_id)
        if match_adj:
            adj_str = match_adj.group(1)
            pivot_zi_id = zone_sm.zi.id_
            return resolv_adj_str(adj_str)
        
        # Has to be done here as regexp is greedy
        # Try nnn[smhdw] 
        match_unit = self._zi_unit_re.search(zi_id)
        if match_unit:
            amount = match_unit.group(1)
            unit = match_unit.group(2)
            try:
                amount = float(amount)
            except ValueError:
                raise ZiIdTimeAmountSyntaxError(zi_id)
            # Use DB server as time base
            now = db_clock_time(db_session)
            try:
                if unit == 's':
                    delta_time = datetime.timedelta(seconds=amount)
                elif unit == 'm':
                    delta_time = datetime.timedelta(minutes=amount)
                elif unit == 'h':
                    delta_time = datetime.timedelta(hours=amount)
                elif unit == 'd':
                    delta_time = datetime.timedelta(days=amount)
                elif unit == 'w':
                    delta_time = datetime.timedelta(weeks=amount)
                else:
                    raise ZiIdTimeUnitSyntaxError(zi_id)
            except ValueError:
                raise ZiIdTimeAmountSyntaxError(zi_id)
            target_ctime = now - delta_time
            query = query.filter(ZoneInstance.ctime <= target_ctime)\
                        .order_by(ZoneInstance.ctime.desc()).limit(1)
            try:
                result = query.one()
            except NoResultFound:
                result = None
            return result
        
        # Try DD/MM/YYYY,hh:mm
        match_ddmmyyyy_hhmm = self._zi_ddmmyyyy_hhmm_re.search(zi_id)
        if match_ddmmyyyy_hhmm:
            return do_year_date_time(match_ddmmyyyy_hhmm,
                                    ZiIdDdMmYyyySyntaxError,
                                    iso_format_date=False)
        # Try YYYY-MM-DD,hh:mm
        match_iso_date_hhmm = self._zi_iso_date_hhmm_re.search(zi_id)
        if match_iso_date_hhmm:
            return do_year_date_time(match_iso_date_hhmm,
                                    ZiIdIsoDateSyntaxError,
                                    iso_format_date=True)
        # Try DD/MM/YYYY
        match_ddmmyyyy = self._zi_ddmmyyyy_re.search(zi_id)
        if match_ddmmyyyy:
            return do_year_date_time(match_ddmmyyyy, ZiIdDdMmYyyySyntaxError,
                                    iso_format_date=False)
        # Try YYYY-MM-DD
        match_iso_date = self._zi_iso_date_re.search(zi_id)
        if match_iso_date:
            return do_year_date_time(match_iso_date, ZiIdIsoDateSyntaxError,
                                    iso_format_date=True)

        # Try DD/MM
        match_ddslashmm = self._zi_ddslashmm_re.search(zi_id)
        if match_ddslashmm:
            day = match_ddslashmm.group(1)
            month = match_ddslashmm.group(2)
            try:
                day = int(day)
            except ValueError:
                raise ZiIdDdSlashMmSyntaxError(zi_id)
            try:
                month = int(month)
            except ValueError:
                raise ZiIdDdSlashMmSyntaxError(zi_id)
            now = db_clock_time(db_session)
            midnight = datetime.time(0, 0, 0, tzinfo=now.tzinfo)
            now_year = now.year
            last_year = now_year - 1
            try:
                target_date = datetime.date(now_year, month, day)
                target_ctime = datetime.datetime.combine(target_date, midnight)
                if target_ctime > now:
                    target_date = datetime.date(last_year, month, day)
                    target_ctime = datetime.datetime.combine(target_date,
                            midnight)
            except ValueError:
                raise ZiIdDdSlashMmSyntaxError(zi_id)
            return ctime_query(target_ctime)

        # Try HH:MM
        match_hhmm = self._zi_hhmm_re.search(zi_id)
        if match_hhmm:
            match_args = match_hhmm.groups()
            hour = match_args[0]
            minute = match_args[1]
            try:
                hour = int(hour)
            except ValueError:
                raise ZiIdHhMmSyntaxError(zi_id)
            try:
                minute = int(minute)
            except ValueError:
                raise ZiIdHhMmSyntaxError(zi_id)
            # Process AM/PM
            if len(match_args) > 2 and match_args[2]:
                am_pm = match_args[2].lower()
                if (am_pm.startswith('p') and hour < 12):
                    hour += 12
            # Use DB server as time base
            now = db_clock_time(db_session)
            now_date = now.date()
            try:
                target_time = datetime.time(hour, minute, tzinfo=now.tzinfo)
                yesterday_date = now_date - datetime.timedelta(days=1)
                target_ctime = datetime.datetime.combine(now_date, target_time)
                if target_ctime > now:
                    # Use yesterday
                    target_ctime = datetime.datetime.combine(yesterday_date, 
                            target_time)
            except ValueError:
                raise ZiIdHhMmSyntaxError(zi_id) 
            return ctime_query(target_ctime)
    
        # Try nnn+++/---/+n/-n
        match_int_adj = self._zi_int_adj_re.search(zi_id)
        if match_int_adj:
            pivot_zi_id = match_int_adj.group(1)
            adj_str = match_int_adj.group(2)
            return resolv_adj_str(adj_str)
        
        # Can't understand whats been given
        raise ZiIdSyntaxError(zi_id)

    def list_zi(self, name):
        """
        Given a zone name, return all its zis briefly, 
        fully showing the currently active one.
        """
        # Deal with SA auto-BEGIN - want fresh transaction to see fresh data
        self._begin_op()
        zone_sm = self._get_zone_sm(name)
        result = zone_sm.to_engine(time_format=self.time_format)
        if zone_sm.zi:
            result['zi'] = zone_sm.zi.to_engine(time_format=self.time_format)
        # Ineffecient but portable code:
        #result['all_zis'] = list(zone_sm.all_zis)
        #result['all_zis'].sort(key=(lambda zi : zi.mtime))
        #result['all_zis'] = [zi.to_engine_brief(time_format=self.time_format)
        #                        for zi in zone_sm.all_zis]
        # Efficient code:
        result['all_zis'] = [zi.to_engine_brief(time_format=self.time_format)
                        for zi in zone_sm.all_zis.order_by(ZoneInstance.ctime)]
        return result

    def _get_comment(self, comment_id):
        """
        Give a comment ID, get the contents of the comment
        """
        db_session = self.db_session
        result = db_session.query(RRComment) \
                .filter(RRComment.id_ == comment_id).one()
        if result:
            rr_comment = result[0].comment
        else:
            rr_comment = None
        return rr_comment

    def _show_zone(self, zone_sm, zi_id=None, all_rrs=False):
        """
        Given a zone_sm, return all the values stored in its ZoneSM
        record, current zi, RRs, and comments
        """
        if not zone_sm:
            return {}
        result = zone_sm.to_engine(time_format=self.time_format)
        if self.sectag.sectag == settings['admin_sectag']:
            result['sectags'] = zone_sm.list_sectags(self.db_session)
        # This is a bit rabbit-pathed, but it works...
        zi = self._resolv_zi_id(zone_sm, zi_id)
        if not zi:
            raise ZiNotFound(zone_sm.name, zi_id)
        result['zi'] = zi.to_data(self.time_format,
                                        zone_sm.use_apex_ns, all_rrs)
        # Note alternative code up in list_zi() for different relN loading
        # strategy
        result['all_zis'] = [zi.to_engine_brief(time_format=self.time_format)
                        for zi in zone_sm.all_zis.order_by(ZoneInstance.ctime)]
        return result

    def show_zone(self, name, zi_id=None):
        """
        Given a zone name, return all the values stored in its ZoneSM
        record, current zi, RRs, and comments
        """
        # Deal with SA auto-BEGIN - want fresh transaction to see fresh data
        self._begin_op()
        zone_sm = self._get_zone_sm(name)
        return self._show_zone(zone_sm, zi_id)
    
    def show_zone_byid(self, zone_id, zi_id=None):
        """
        Given a zone id, return all the values stored in its ZoneSM
        record, current zi, RRs, and comments
        """
        # Deal with SA auto-BEGIN - want fresh transaction to see fresh data
        self._begin_op()
        zone_sm = self._get_zone_sm_byid(zone_id)
        return self._show_zone(zone_sm, zi_id)

    def show_zone_text(self, name, zi_id=None, all_rrs=True):
        """
        Given a zone name and optional zi_id, return the ZI as zone file text
        """
        # Deal with SA auto-BEGIN - want fresh transaction to see fresh data
        self._begin_op()
        result = {}
        zone_sm = self._get_zone_sm(name)
        data_result = self._show_zone(zone_sm, zi_id, all_rrs=all_rrs)
        result['zi_text'] = data_to_bind(data_result['zi'], 
                                name=data_result['name'], 
                                reference=data_result.get('reference'))
        result['name'] = data_result['name']
        result['zi_id'] = data_result['zi']['zi_id']
        result['zi_ctime'] = data_result['zi']['ctime']
        result['zi_mtime'] = data_result['zi']['mtime']
        result['zi_ptime'] = data_result['zi']['ptime']
        result['soa_serial'] = data_result['zi']['soa_serial']
        result['zone_id'] = data_result['zone_id']
        return result

    def show_zi(self, name, zi_id=None):
        """
        Given a domain name and optionally a zi_id, return all values
        stored in ZoneInstance record
        """
        # Deal with SA auto-BEGIN - want fresh transaction to see fresh data
        self._begin_op()
        zone_sm = self._get_zone_sm(name)
        zi = self._resolv_zi_id(zone_sm, zi_id)
        if not zi:
            raise ZiNotFound(name, zi_id)
        result = zi.to_engine(time_format=self.time_format)
        return result

    def show_zi_byid(self, zi_id):
        """
        Given a zi_id, return all values stored in ZoneInstance record
        """
        # Deal with SA auto-BEGIN - want fresh transaction to see fresh data
        self._begin_op()
        zi = self._get_zi(zi_id)
        result = zi.to_engine(time_format=self.time_format)
        return result

    def _edit_zone(self, name, login_id, zi_id=None, all_rrs=False,
                    admin_privilege=False):
        """
        Backend for zone editing.

        Start editing a zone, by returning editing data

        If zone has edit locking enabled, change state and obtain a
        token
        """
        # Deal with SA auto-BEGIN - want fresh transaction to see fresh data
        self._begin_op()
        locked_by = self._make_change_by(login_id)
        zone_sm = self._get_zone_sm(name)
        # Privilege check for no apex zones - admin only 
        if not zone_sm.use_apex_ns and not admin_privilege:
            raise ZoneAdminPrivilegeNeeded(name)
        edit_lock_token = None
        if zone_sm.edit_lock:
            # This is where we synchronously call the Zone_sm state
            # machine Have to obtain lock before getting current data
            lock_results = exec_zonesm(zone_sm, ZoneSMEdit, EditLockFailure,
                    locked_by=locked_by)
            edit_lock_token = lock_results['edit_lock_token']
       
        # All locking now done, get zone data and return it!
        try:
            zone_zi_data = self._show_zone(zone_sm, zi_id, all_rrs)
        except ZoneNotFound:
            # If fail to obtain data release edit lock
            if zone_sm.state == ZLSTATE_EDIT_LOCK:
                #Cancel Edit lock
                exec_zonesm(zone_sm, ZoneSMEditExit, 
                            edit_lock_token=edit_lock_token)
            raise
        # return with THE HIDDEN TREASURE
        return zone_zi_data, edit_lock_token

    def edit_zone(self, name, login_id, zi_id=None):
        """
        Start editing a zone, by returning editing data

        If zone has edit locking enabled, change state and obtain a
        token
        """
        return self._edit_zone(name, login_id, zi_id)
    
    def edit_zone_admin(self, name, login_id, zi_id=None):
        """
        Start editing a zone, by returning editing data

        If zone has edit locking enabled, change state and obtain a
        token
        """
        return self._edit_zone(name, login_id, zi_id, all_rrs=True,
                                admin_privilege=True)

    def tickle_editlock(self, name, edit_lock_token=None):
        """
        Tickle the edit_lock timeout event
        """
        # Deal with SA auto-BEGIN - want fresh transaction to see fresh data
        self._begin_op()
        zone_sm = self._get_zone_sm(name)

        exec_zonesm(zone_sm, ZoneSMEditLockTickle, 
                    TickleEditLockFailure,
                    edit_lock_token=edit_lock_token)
        return True

    def cancel_edit_zone(self, name, edit_lock_token=None):
        """
        Operation to cancel an edit locked session
        """
        # Deal with SA auto-BEGIN - want fresh transaction to see fresh data
        self._begin_op()
        zone_sm = self._get_zone_sm(name)
        cancel_results = exec_zonesm(zone_sm, ZoneSMEditExit,
                                    CancelEditLockFailure,
                                    edit_lock_token=edit_lock_token)
        return True

    def _update_zone(self, name, zi_data, login_id, edit_lock_token=None,
            normalize_ttls=False, admin_privilege=False,
            helpdesk_privilege=False):
        """
        Backend for updating the zone by adding a new ZI,
        and emitting a publish event
        """
        # Deal with SA auto-BEGIN - want fresh transaction to see fresh data
        self._begin_op()
        change_by = self._make_change_by(login_id)
        zone_sm = self._get_zone_sm(name, exact_network=True)
        # Privilege check for no apex zones - admin only 
        if not zone_sm.use_apex_ns and not admin_privilege:
            raise ZoneAdminPrivilegeNeeded(name)
        # Save data
        zi, auto_ptr_data = self._data_to_zi(name, zi_data, change_by,
                                normalize_ttls, 
                                admin_privilege, helpdesk_privilege)
        # put zi in place, issue appropriate zone SM event
        if not zone_sm.edit_lock:
            exec_zonesm(zone_sm, ZoneSMEditSavedNoLock,
                                zi_id=zi.id_)

            # Do auto_ptr_data operation here.
            self._queue_auto_ptr_data(auto_ptr_data)
            return True

        try:
            exec_zonesm(zone_sm, ZoneSMEditSaved,
                                UpdateZoneFailure,
                                zi_id=zi.id_,
                                edit_lock_token=edit_lock_token)
        except UpdateZoneFailure as exc:
            # Remove zi as we don't want to keep it around
            # - Obviates edit locking in the first place.
            self.db_session.delete(zi)
            self.db_session.commit()
            raise
        
        # Do auto_ptr_data operation here.
        self._queue_auto_ptr_data(auto_ptr_data)
        return True
        
    def _update_zone_text(self, name, zi_text, login_id, edit_lock_token=None,
            normalize_ttls=False, admin_privilege=False,
            helpdesk_privilege=False):
        """
        Backend for updating the zone by adding a new ZI as a text blob,
        and emitting a publish event
        """
        zi_data, origin_name, update_type, zone_reference \
                                = self._parse_zi_text(name, zi_text)
        # Use normalize_ttls with imported data to stop surprises
        results = self._update_zone(name=name, login_id=login_id, 
                            zi_data=zi_data, edit_lock_token=edit_lock_token,
                            normalize_ttls=normalize_ttls,
                            admin_privilege=admin_privilege,
                            helpdesk_privilege=helpdesk_privilege)
        return results

    def update_zone_admin(self, name, zi_data, login_id, edit_lock_token=None, 
                            normalize_ttls=False):
        """
        Update a zone with admin privilege
        """
        return self._update_zone(name, zi_data, login_id, edit_lock_token, 
                    normalize_ttls, admin_privilege=True)

    def update_zone_text_admin(self, name, zi_text, login_id, 
                            edit_lock_token=None, normalize_ttls=False):
        """
        Update a zone with admin privilege
        """
        return self._update_zone_text(name, zi_text, login_id, 
                            edit_lock_token, normalize_ttls, 
                            admin_privilege=True)

    def _find_src_zi(self, src_name, src_zi_id, admin_privilege):
        """
        Find a src_zi, src_zone_sm given a name and zi_id

        Common peice of code between _create_zone and _copy_zi
        """
        db_session = self.db_session
        src_zone_sm = None
        src_zi = None
        if src_name:
            src_zone_sm = self._get_zone_sm(src_name)
            src_zi = self._resolv_zi_id(src_zone_sm, src_zi_id)
            if not src_zi:
                raise ZiNotFound(src_zone_sm.name, src_zi_id)
        elif src_zi_id and admin_privilege:
            src_zi = self._resolv_zi_id(None, src_zi_id,
                            specific_zi_id=True)
            if not src_zi:
                raise ZiNotFound('*', src_zi_id)
        return src_zone_sm, src_zi


    def _copy_src_zi(self, src_zi, zone_sm, change_by,
                    preserve_time_stamps=False):
        """
        Given a src_zi, copy it

        Common peice of code between _create_zone and _copy_zi
        """
        db_session = self.db_session
        if preserve_time_stamps:
            src_ctime = src_zi.ctime
            src_mtime = src_zi.mtime
        zi = src_zi.copy(db_session, change_by)
        auto_ptr_data = zi.get_auto_ptr_data(zone_sm)
        # Tie to zone
        zi.zone = zone_sm
        zone_sm.all_zis.append(zi)
        db_session.flush()
        # Update apex if needed
        zi.update_apex(db_session)
        # Update Zone TTLs for clean initialisation
        zi.update_zone_ttls()
        # Make sure SOA serial number is fresh
        new_soa_serial = new_soa_serial_no(zi.soa_serial, zone_sm.name)
        zi.update_soa_serial(new_soa_serial)
        if preserve_time_stamps:
            zi.ctime = src_ctime
            zi.mtime = src_mtime
        db_session.flush()
        return zi, auto_ptr_data

    def _create_zone(self, name, zi_data, login_id,
                        use_apex_ns, edit_lock, auto_dnssec, nsec3, 
                        inc_updates,
                        reference=None, sg_name=None, sectags=None,
                        batch_load=False, src_name=None, src_zi_id=None,
                        admin_privilege=False,
                        helpdesk_privilege=False):
        """
        Given a name, create a zone

        Currently just creates a row in the sm_zone table, as well as
        initial zi (provided or default), leaving  zone_sm.state as UNCONFIG
        """

        def check_parent_domains(name):
            """
            Handle all checks for when creating sub domain

            ie - only allow sub domain creation for like references etc
            """
            nonlocal reference

            # Check if sub domain exists
            parent_name_list = name.split('.')[1:]
            while (len(parent_name_list) > 1):
                parent_name  = '.'.join(parent_name_list)
                parent_name_list = parent_name_list[1:]
                try:
                    parent_zone_sm = self._get_zone_sm(parent_name,
                            check_sectag=False, exact_network=True)
                except ZoneNotFound:
                    continue
                parent_zone_ref = parent_zone_sm.reference
                if self.sectag.sectag == settings['admin_sectag']:
                    # admin can do anything - creating a sub domain 
                    # with any reference defaulting to that of parent
                    if not reference:
                        reference = parent_zone_ref.reference \
                            if hasattr(parent_zone_ref, 'reference') \
                                and parent_zone_ref.reference \
                            else None
                    return
                if not reference:
                    reference = zone_cfg.get_row_exc(db_session, 'default_ref')
                ref_obj = new_reference(db_session, reference,
                        return_existing=True)
                if parent_zone_ref and ref_obj != parent_zone_ref:
                    raise ZoneExists(name)
                return
            return

        self._begin_op()
        db_session = self.db_session
       
        # Login ID must be checked and processed
        change_by = self._make_change_by(login_id)

        # If given source information for copying into creation ZI, check
        # it out.
        src_zone_sm, src_zi = self._find_src_zi(src_name, src_zi_id, 
                                admin_privilege)

        try:
            # No point in hiding existence of zone if asked directly with 
            # name when creating a zone.
            zone_sm = self._get_zone_sm(name, check_sectag=False, 
                            exact_network=True)
            # reached the end of the road...
            raise ZoneExists(name)
        except ZoneNotFound:
            # Inverted exception
            pass

        # Check name syntax and convert networks to valid reverse domain names
        reverse_network = None
        if name.find('/') > -1:
            result = zone_name_from_network(name)
            if not result:
                raise InvalidDomainName(name)
            rev_name, rev_network = result
            reverse_network = new_reverse_network(db_session, rev_network)
            name = rev_name
            inc_updates = True if inc_updates == None else inc_updates
        elif (name.lower().endswith('ip6.arpa.') 
                or name.lower().endswith('in-addr.arpa.')):
            raise ReverseNamesNotAccepted(name)
        elif not is_inet_domain(name):
            raise InvalidDomainName(name)

        # Check parent domains when creating a sub domain
        check_parent_domains(name)

        # Set reference if copying and none given.
        # Parent domains will override this
        if src_zone_sm and not reference:
            if src_zone_sm.reference and src_zone_sm.reference.reference:
                reference = src_zone_sm.reference.reference

        # Check that the security tag exists
        sectag = self.sectag
        if not sectag in list_all_sectags(db_session):
            raise ZoneSecTagDoesNotExist(sectag.sectag)

        # If copying zone, set zone flags from src if not given
        if src_zone_sm:
            if use_apex_ns is None:
                use_apex_ns = src_zone_sm.use_apex_ns
            if edit_lock is None:
                edit_lock = src_zone_sm.edit_lock
            if auto_dnssec is None:
                auto_dnssec = src_zone_sm.auto_dnssec
            if nsec3 is None:
                nsec3 = src_zone_sm.nsec3
            if inc_updates is None:
                inc_updates = src_zone_sm.inc_updates

        # create the zone
        zone_sm = new_zone(db_session, DynDNSZoneSM, name=name, 
                    use_apex_ns=use_apex_ns, edit_lock=edit_lock,
                    auto_dnssec=auto_dnssec, nsec3=nsec3,
                    inc_updates=inc_updates, sectag=self.sectag,
                    sg_name=sg_name, reference=reference)
        # Add extra sectags
        if sectags:
            if self.sectag.sectag == settings['admin_sectag']:
                self.replace_zone_sectags(name, sectags)
            else:
                raise SecTagPermissionDenied(self.sectag.sectag)
        
        # If Admin and copying, copy sectags from source zone
        if self.sectag.sectag == settings['admin_sectag']:
            if src_zone_sm:
                zone_sm.copy_zone_sectags(db_session, src_zone_sm)

        # Fill out zi
        if src_zi:
            zi, auto_ptr_data = self._copy_src_zi(src_zi, zone_sm, change_by)
        elif zi_data:
            zi, auto_ptr_data = self._data_to_zi(name, zi_data,
                                change_by=change_by,
                                admin_privilege=admin_privilege, 
                                helpdesk_privilege=helpdesk_privilege,
                                normalize_ttls=True)
            # Set new SOA serial if it is old.  This is for load_zone(s), and
            # new incoming domains
            new_soa_serial = new_soa_serial_no(zi.soa_serial, name)
            zi.update_soa_serial(new_soa_serial)
        else:
            zi = new_zone_zi(db_session, zone_sm, change_by)
            auto_ptr_data = None

        zone_sm.soa_serial = zi.soa_serial
        # Add reverse network if that exists
        if reverse_network:
            zone_sm.reverse_network = reverse_network
        # Get commands going with working backend first
        if (batch_load and not zone_sm.auto_dnssec):
            exec_zonesm(zone_sm, ZoneSMDoBatchConfig, zi_id=zi.id_)
        else:
            exec_zonesm(zone_sm, ZoneSMDoConfig, zi_id=zi.id_)

        # Do auto_ptr_data operation here.
        self._queue_auto_ptr_data(auto_ptr_data)
        # Commit everything.
        self._finish_op()
        return True

    def create_zone_admin(self, name, login_id, zi_data=None, 
                use_apex_ns=None, edit_lock=None, auto_dnssec=None,
                nsec3=None, inc_updates=None, reference=None, sg_name=None,
                sectags=None):
        """
        Create a zone with admin privilege
        """
        return self._create_zone(name, zi_data, login_id, use_apex_ns,
                edit_lock, auto_dnssec, nsec3, inc_updates, 
                reference, sg_name, sectags, admin_privilege=True)

    def copy_zone_admin(self, src_name, name, login_id, zi_id=None, 
                use_apex_ns=None, edit_lock=None, auto_dnssec=None,
                nsec3=None, inc_updates=None, reference=None, sg_name=None,
                sectags=None):
        """
        Create a zone with admin privilege
        """
        return self._create_zone(name, src_name=src_name, src_zi_id=zi_id,
                use_apex_ns=use_apex_ns, edit_lock=edit_lock, 
                auto_dnssec=auto_dnssec, nsec3=nsec3, inc_updates=inc_updates, 
                reference=reference, sg_name=sg_name, sectags=sectags,
                login_id=login_id, zi_data=None, admin_privilege=True)

    def destroy_zone(self, zone_id):
        """
        Destroy a zone backend
        """
        self._begin_op()
        zone_sm = self._get_zone_sm_byid(zone_id)
        
        if not zone_sm.is_deleted():
            raise ZoneNotDeleted(zone_sm.name)

        # Delete the zone
        # Database integrity constraints/triggers will do all the rest...
        exec_zonesm(zone_sm, ZoneSMDoDestroy, ZoneFilesStillExist)
        self._finish_op()
        return True

    def delete_zone(self, name):
        """
        Delete a zone backend
        """
        self._begin_op()
        zone_sm = self._get_zone_sm(name, exact_network=True)
        exec_zonesm(zone_sm, ZoneSMDelete, ZoneBeingCreated)
        self._finish_op()        

    def undelete_zone(self, zone_id):
        """
        Delete a zone backend
        """
        self._begin_op()
        zone_sm = self._get_zone_sm_byid(zone_id)
        exec_zonesm(zone_sm, ZoneSMUndelete, ActiveZoneExists)
        self._finish_op()

    def copy_zi(self, src_name, name, login_id, zi_id=None):
        """
        Copy a zi from src_zone to destination zone
        """
        self._begin_op()
        change_by = self._make_change_by(login_id)
        src_zone_sm, src_zi = self._find_src_zi(src_name, zi_id, 
                                admin_privilege=False)
        zone_sm = self._get_zone_sm(name)
        self._copy_src_zi(src_zi, zone_sm, change_by, 
                        preserve_time_stamps=True)
        self._finish_op()

    def delete_zi(self, name, zi_id):
        """
        Given a zone name and zi_id, delete the zi_id
        """
        self._begin_op()
        zone_sm = self._get_zone_sm(name, exact_network=True)
        
        if zone_sm.zi_id == zi_id:
            raise ZiInUse(name, zi_id)
        zi = self._resolv_zi_id(zone_sm, zi_id, specific_zi_id=True)
        if not zi:
            raise ZiNotFound(name, zi_id)
        self.db_session.delete(zi)
        self._finish_op()

    def _parse_zi_text(self, name, zi_text):
        """
        Backend function to parse zi_text and trap/translate PyParsing
        exceptions.
        """
        zone_stringio = StringIO(initial_value=zi_text)
        try:
            return bind_to_data(zone_stringio, name)
        except ParseBaseException as exc:
            raise ZiTextParseError(name, exc)

    def _load_zone(self, name, zi_text, login_id,
                        use_apex_ns, edit_lock, auto_dnssec, nsec3, 
                        inc_updates,
                        reference=None, sg_name=None, sectags=None,
                        admin_privilege=False,
                        helpdesk_privilege=False):
        """
        Load a zone from a zi_text blob. Backend.
        """
        zi_data, origin_name, update_type, zone_reference \
                        = self._parse_zi_text(name, zi_text)
        if not reference:
            reference = zone_reference
        results = self._create_zone(name, zi_data, login_id,
                        use_apex_ns, edit_lock, auto_dnssec, nsec3, 
                        inc_updates,
                        reference, sg_name, sectags,
                        admin_privilege=admin_privilege,
                        helpdesk_privilege=helpdesk_privilege)
        return results

    def _load_zi(self, name, zi_text, login_id, admin_privilege=False,
            helpdesk_privilege=False):
        """
        Load a ZI into a zone. Backend.
        """
        zone_sm_data, edit_lock_token = self._edit_zone(name=name,
                                login_id=login_id, 
                                admin_privilege=admin_privilege)
        zi_data, origin_name, update_type, zone_reference \
                                = self._parse_zi_text(name, zi_text)
        # Use normalize_ttls with imported data to stop surprises
        load_results = self._update_zone(name=name, login_id=login_id, 
                            zi_data=zi_data, edit_lock_token=edit_lock_token,
                            normalize_ttls=True,
                            admin_privilege=admin_privilege,
                            helpdesk_privilege=helpdesk_privilege)
        return load_results

    def _set_zone(self, name, **kwargs):
        """
        Set the settable attributes on a zone.  This call also issues
        an event to update the zone.
        """
        for arg in kwargs:
            if arg not in ('use_apex_ns', 'edit_lock', 'auto_dnssec', 'nsec3',
                            'inc_updates'):
                raise InvalidParamsJsonRpcError("Argument '%s' not supported."
                                                    % arg)

        self._begin_op()
        zone_sm = self._get_zone_sm(name, exact_network=True)
       
        if 'use_apex_ns' in kwargs:
            use_apex_ns = kwargs['use_apex_ns']
            if use_apex_ns == None:
                use_apex_ns = zone_cfg.get_key(self.db_session, 'use_apex_ns')
            if use_apex_ns == True:
                if not zone_sm.use_apex_ns:
                    zone_sm.use_apex_ns = True
                    create_event(ZoneSMUpdate, commit=True,
                            signal_queue_daemon=True, 
                            sm_id=zone_sm.id_, zone_id=zone_sm.id_,
                            name=zone_sm.name)
            elif use_apex_ns == False:
                if zone_sm.use_apex_ns:
                    zone_sm.use_apex_ns = False
                    create_event(ZoneSMUpdate, commit=True,
                            signal_queue_daemon=True, 
                            sm_id=zone_sm.id_, zone_id=zone_sm.id_,
                            name=zone_sm.name)
            else:
                assert(False)

        if 'edit_lock' in kwargs:
            edit_lock = kwargs['edit_lock']
            if edit_lock == None:
                edit_lock = zone_cfg.get_key(self.db_session, 'edit_lock')
            if edit_lock == True:
                zone_sm.edit_lock = True
            elif edit_lock == False:
                zone_sm.edit_lock = False
            elif edit_lock == None:
                pass
            else:
                assert(False)

        if 'inc_updates' in kwargs:
            inc_updates = kwargs['inc_updates']
            if inc_updates == None:
                inc_updates = zone_cfg.get_key(self.db_session, 'inc_updates')
            if inc_updates == True:
                zone_sm.inc_updates = True
            elif inc_updates == False:
                zone_sm.inc_updates = False
            elif inc_updates == None:
                pass
            else:
                assert(False)

        if 'auto_dnssec' in kwargs:
            auto_dnssec = kwargs['auto_dnssec']
            if auto_dnssec == None:
                auto_dnssec = zone_cfg.get_key(self.db_session, 'auto_dnssec')
            if auto_dnssec == True:
                if not zone_sm.auto_dnssec:
                    zone_sm.auto_dnssec = True
                    exec_zonesm(zone_sm, ZoneSMDoReconfig)
            elif auto_dnssec == False:
                if zone_sm.auto_dnssec:
                    zone_sm.auto_dnssec = False
                    exec_zonesm(zone_sm, ZoneSMDoReconfig)
            elif auto_dnssec == None:
                pass
            else:
                assert(False)

        if 'nsec3' in kwargs:
            nsec3 = kwargs['nsec3']
            if nsec3 == None:
                nsec3 = zone_cfg.get_key(self.db_session, 'nsec3')
            if nsec3 == True:
                if not zone_sm.nsec3:
                    zone_sm.nsec3 = True
                    if zone_sm.auto_dnssec:
                        exec_zonesm(zone_sm, ZoneSMDoReconfig)
            elif nsec3 == False:
                if zone_sm.nsec3:
                    zone_sm.nsec3 = False
                    if zone_sm.auto_dnssec:
                        exec_zonesm(zone_sm, ZoneSMDoReconfig)
            elif nsec3 == None:
                pass
            else:
                assert(False)
        self._finish_op()
        return

    def set_zone_admin(self, name, **kwargs):
        return self._set_zone(name, **kwargs)

    def show_sectags(self):
        """
        Return all security tags as JSON
        """
        if self.sectag.sectag != settings['admin_sectag']:
            raise SecTagPermissionDenied(self.sectag.sectag)
        self._begin_op()
        result = []
        all_sectags = list_all_sectags(self.db_session)
        if not len(all_sectags):
            raise NoSecTagsExist()
        for sectag in all_sectags:
            result.append(sectag.to_engine(self.time_format))
        self._finish_op()
        return result

    def show_zone_sectags(self, name):
        """
        Return all the sectags configured for a zone
        """
        if self.sectag.sectag != settings['admin_sectag']:
            raise SecTagPermissionDenied(self.sectag.sectag)
        self._begin_op()
        zone_sm = self._get_zone_sm(name)
        result = zone_sm.list_sectags(self.db_session)
        if not result:
            raise NoZoneSecTagsFound(name)
        self._finish_op()
        return result

    def add_zone_sectag(self, name, sectag_label):
        """
        Add a sectag to a zone
        """
        if self.sectag.sectag != settings['admin_sectag']:
            raise SecTagPermissionDenied(self.sectag.sectag)
        self._begin_op()
        zone_sm = self._get_zone_sm(name, exact_network=True)
        sectag = ZoneSecTag(sectag_label)
        if sectag not in list_all_sectags(self.db_session):
            raise ZoneSecTagDoesNotExist(sectag_label)
        result = zone_sm.add_sectag(self.db_session, sectag)
        self._finish_op()
    
    def delete_zone_sectag(self, name, sectag_label):
        """
        Add a sectag to a zone
        """
        if self.sectag.sectag != settings['admin_sectag']:
            raise SecTagPermissionDenied(self.sectag.sectag)
        self._begin_op()
        zone_sm = self._get_zone_sm(name, exact_network=True)
        sectag = ZoneSecTag(sectag_label)
        if sectag not in list_all_sectags(self.db_session):
            raise ZoneSecTagDoesNotExist(sectag_label)
        result = zone_sm.remove_sectag(self.db_session, sectag)
        self._finish_op()

    def replace_zone_sectags(self, name, sectag_labels):
        """
        Replace all sectags for given zone
        """
        if self.sectag.sectag != settings['admin_sectag']:
            raise SecTagPermissionDenied(self.sectag.sectag)
        self._begin_op()
        zone_sm = self._get_zone_sm(name, exact_network=True)
        sectag_list = []
        all_sectags = list_all_sectags(self.db_session)
        for sectag_thing in sectag_labels:
            try:
                sectag = ZoneSecTag(sectag_thing['sectag_label'])
            except (TypeError, IndexError):
                raise InvalidParamsJsonRpcError('Sectag list format invalid.')
            if sectag not in all_sectags:
                raise ZoneSecTagDoesNotExist(sectag_thing['sectag_label'])
            sectag_list.append(sectag)
        result = zone_sm.replace_all_sectags(self.db_session, *sectag_list)
        self._finish_op()
   
    def enable_zone(self, name):
        """
        Enable a zone
        """
        self._begin_op()
        zone_sm = self._get_zone_sm(name, exact_network=True)
        exec_zonesm(zone_sm, ZoneSMEnable)
        self._finish_op()


    def disable_zone(self, name):
        """
        Disable a zone
        """
        self._begin_op()
        zone_sm = self._get_zone_sm(name, exact_network=True)
        exec_zonesm(zone_sm, ZoneSMDisable)
        self._finish_op()

    def show_mastersm(self):
        """
        Show the MasterSM
        """
        self._begin_op()
        result = show_master_sm(self.db_session, time_format=self.time_format)
        self._finish_op()
        return result

    def sign_zone(self, name):
        """
        Schedule a zone for signing event
        """
        self._begin_op()
        zone_sm = self._get_zone_sm(name, exact_network=True)
        if not zone_sm.auto_dnssec:
            raise ZoneNotDnssecEnabled(name)
        zone_sm_dnssec_schedule(self.db_session, zone_sm, 'sign')
        self._finish_op()
    
    def loadkeys_zone(self, name):
        """
        Schedule a zone key loading event
        """
        self._begin_op()
        zone_sm = self._get_zone_sm(name, exact_network=True)
        if not zone_sm.auto_dnssec:
            raise ZoneNotDnssecEnabled(name)
        zone_sm_dnssec_schedule(self.db_session, zone_sm, 'loadkeys')
        self._finish_op()

    def reset_zone(self, name, zi_id=None):
        """
        Schedule a zone reset event
        """
        self._begin_op()
        zone_sm = self._get_zone_sm(name, exact_network=True)
        reset_args = {}
        if zi_id:
            zi = self._resolv_zi_id(zone_sm, zi_id, specific_zi_id=False)
            if not zi:
                raise ZiNotFound(zone_sm.name, zi_id)
            reset_args['zi_id'] = zi.id_
        results = exec_zonesm(zone_sm, ZoneSMDoReset, **reset_args)
        self._finish_op()

    def refresh_zone(self, name, zi_id=None):
        """
        Refresh a zone by issuing an update.
        """
        self._begin_op()
        zone_sm = self._get_zone_sm(name)
        refresh_args = {}
        if zi_id:
            zi = self._resolv_zi_id(zone_sm, zi_id, specific_zi_id=False)
            if not zi:
                raise ZiNotFound(zone_sm.name, zi_id)
            refresh_args['zi_id'] = zi.id_
            results = exec_zonesm(zone_sm, ZoneSMDoRefresh, 
                    exception_type=UpdateZoneFailure, **refresh_args)
        else:
            results = exec_zonesm(zone_sm, ZoneSMDoRefresh, **refresh_args)
        self._finish_op()

    def poke_zone_set_serial(self, name, soa_serial=None,
                                            force_soa_serial_update=False):
        """
        Set zone SOA serial number to given value if possible
        """
        return self._poke_zone(name, soa_serial=soa_serial,
                            force_soa_serial_update=force_soa_serial_update)

    def poke_zone_wrap_serial(self, name):
        """
        Wrap current zone SOA serial number
        """
        return self._poke_zone(name, wrap_soa_serial=True)

    def _poke_zone(self, name, soa_serial=None,
                                wrap_soa_serial=False,
                                force_soa_serial_update=False):
        """
        Manipulate a zone's serial number on the DNS servers via update.
        """
        self._begin_op()
        zone_sm = self._get_zone_sm(name)
        if zone_sm.state != ZSTATE_PUBLISHED:
            raise ZoneNotPublished(name)
        # If candidate serial given, test it
        if soa_serial:
            # Check that incoming argument is an integer
            if not isinstance(soa_serial, int):
                raise SOASerialTypeError(name)
            if not ( 0 < soa_serial <= (2**32 -1 )):
                raise SOASerialRangeError(name)
            # Assume that current is previously published SOA serial
            test_soa_serial = new_soa_serial_no(zone_sm.soa_serial, name,
                    candidate=soa_serial)
            if test_soa_serial != soa_serial:
                raise SOASerialCandidateIgnored(name)
        refresh_args = {'candidate_soa_serial': soa_serial, 
                        'wrap_soa_serial': wrap_soa_serial,
                        'force_soa_serial_update': force_soa_serial_update}
        results = exec_zonesm(zone_sm, ZoneSMDoRefresh, 
                exception_type=UpdateZoneFailure, **refresh_args)
        self._finish_op()

    def create_reference(self, reference):
        """
        Create a new reference
        """
        self._begin_op()
        new_reference(self.db_session, reference)
        self._finish_op()

    def delete_reference(self, reference):
        """
        Delete a reference
        """
        self._begin_op()
        del_reference(self.db_session, reference)
        self._finish_op()

    def rename_reference(self, reference, dst_reference):
        """
        Rename a reference
        """
        self._begin_op()
        rename_reference(self.db_session, reference, dst_reference)
        self._finish_op()
    
    def list_reference(self, *references):
        """
        List references
        """
        self._begin_op()
        db_session = self.db_session
        db_query_slice = get_numeric_setting('db_query_slice', int)
        if not references:
            references = '*'
        ref_list = []
        ref_pattern = ' '.join(references)
        references = [x.replace('*', '%') for x in references]
        references = [x.replace('?', '_') for x in references]
        for ref in references:
            query = self.db_session.query(Reference)\
                    .filter(Reference.reference.ilike(ref))\
                    .yield_per(db_query_slice)
            for ref in query:
                ref_list.append(ref.to_engine_brief())
        if not ref_list:
            raise NoReferenceFound('*')
        ref_list = sorted(ref_list, 
                key=lambda reference: reference['reference'].lower())
        self._finish_op()
        return ref_list

    def set_zone_reference(self, name, reference=None):
        """
        Set the reference for a zone
        """
        self._begin_op()
        db_session = self.db_session
        zone_sm = self._get_zone_sm(name, exact_network=True)
        if reference:
            reference = find_reference(db_session, reference)
            reference.set_zone(zone_sm)
        else:
            zone_sm.ref_id = None
        self._finish_op()

    def list_sg(self):
        """
        List all server groups
        """
        self._begin_op()
        sgs = self.db_session.query(ServerGroup).all()
        result = []
        for sg in sgs:
            result.append(sg.to_engine_brief())
        if not result:
            raise NoSgFound('*')
        self._finish_op()
        return result

    def set_zone_sg(self, name, sg_name=None):
        """
        Set the SG a zone is associated with
        """
        self._begin_op()
        db_session = self.db_session
        zone_sm = self._get_zone_sm(name, exact_network=True)
        if not zone_sm.is_disabled():
            raise ZoneNotDisabled(name)
        if not sg_name:
            sg_name = zone_cfg.get_row_exc(db_session, 'default_sg')
        sg = find_sg_byname(db_session, sg_name, raise_exc=True)
        zone_sm.set_sg(sg)
        self._finish_op()

    def set_zone_alt_sg(self, name, sg_name=None):
        """
        Set the alternate SG a zone is associated with
        """
        self._begin_op()
        db_session = self.db_session
        zone_sm = self._get_zone_sm(name, exact_network=True)
        exec_zonesm(zone_sm, ZoneSMDoSetAltSg, ZoneSmFailure, 
                alt_sg_name=sg_name)
        self._finish_op()

    def swap_zone_sg(self, name):
        """
        Swap a live zone's sg over with its alt_sg
        """
        self._begin_op()
        db_session = self.db_session
        zone_sm = self._get_zone_sm(name, exact_network=True)
        if not zone_sm.alt_sg:
            raise ZoneNoAltSgForSwap(name)
        exec_zonesm(zone_sm, ZoneSMDoSgSwap)
        self._finish_op()
        

    def rr_query_db(self, label, name=None, type=None,
        rdata=None, zi_id=None, show_all=False):
        """
        Query the DB for RRs matching the given pattern 
        """
        self._begin_op()
        db_session = self.db_session
        try:
            result = rr_query_db_raw(db_session, label=label, name=name,
                    type_=type, rdata=rdata, include_disabled=show_all,
                    zi_id=zi_id, sectag=self.sectag)
        except ValueError as exc:
            raise RrQueryDomainError(name)
        if result:
            rrs = result.get('rrs')
            if not rrs:
                return None
            result['rrs'] = [rr.to_engine() for rr in rrs]
            result['zone_disabled'] = result['zone_sm'].is_disabled()
            result.pop('zone_sm', None)
        self._finish_op()
        return result

    def _update_rrs(self, name, update_data, update_type, login_id,
                    admin_privilege=False, helpdesk_privilege=False):
        """
        Do Incremental Updates for a zone.  Takes same ZI data format as
        _create_zone().  Will produce a JSON Error if an exception is thrown.
        """
        self._begin_op()
        change_by = self._make_change_by(login_id)
        auto_ptr_data = self._data_to_update(name, update_data, update_type,
                            change_by,
                            admin_privilege=admin_privilege,
                            helpdesk_privilege=helpdesk_privilege)
        # Do auto_ptr_data operation here.
        self._queue_auto_ptr_data(auto_ptr_data)
        # Commit everything.
        self._finish_op()

    def update_rrs_admin(self, name, update_data, update_type, login_id):
        """
        Incremental updates, admin privilege
        """
        return self._update_rrs(name, update_data, update_type, login_id, 
                                admin_privilege=True)

    def refresh_zone_ttl(self, name, zone_ttl=None):
        """
        Refresh a zones TTL by issuing an update.
        """
        self._begin_op()
        zone_sm = self._get_zone_sm(name, exact_network=True)
        if not zone_ttl:
            zone_ttl = zone_cfg.get_row_exc(self.db_session, 'zone_ttl', 
                                    sg_name=zone_sm.sg.name)
        if zone_sm.zi_candidate:
            zone_sm.zi_candidate.update_zone_ttls(zone_ttl=zone_ttl)
        elif zone_sm.zi:
            zone_sm.zi.update_zone_ttls(zone_ttl=zone_ttl)
        else:
            raise ZoneHasNoZi(name)

        results = exec_zonesm(zone_sm, ZoneSMDoRefresh)
        self._finish_op()
    
    def list_pending_events(self):
        """
        List pending events
        """
        self._begin_op()
        db_query_slice = get_numeric_setting('db_query_slice', int)
        db_session = self.db_session
        query = db_session.query(Event).filter(Event.processed == None)\
                .order_by(Event.id_).yield_per(db_query_slice)
        results = []
        for event in query:
            json_event = event.to_engine_brief(time_format=self.time_format)
            results.append(json_event)
        self._finish_op()
        return results

    def _find_sg_byname(self, sg_name):
        """
        Given an sg_id, return the server group
        """
        db_session = self.db_session
        return find_sg_byname(db_session, sg_name, raise_exc=True)
    
    def _show_sg(self, sg):
        """
        Back end - Show the details of an SG
        """
        result = sg.to_engine()
        servers = []
        for server in sg.servers:
            servers.append(server.to_engine_brief()) 
        result['servers'] = servers if servers else None
        self._finish_op()
        return result

    def show_sg(self, sg_name):
        """
        Show the details of an SG
        """
        self._begin_op()
        sg = self._find_sg_byname(sg_name)
        return self._show_sg(sg)

    def show_replica_sg(self):
        """
        Show Master SG - sub call for status display
        """
        self._begin_op()
        db_session = self.db_session
        replica_sg = get_mastersm_replica_sg(db_session)
        if not replica_sg:
            raise NoReplicaSgFound()
        return self._show_sg(replica_sg)

    def list_server(self, *servers, sg_name=None, show_all=True, 
            show_active=False):
        """
        List servers
        """
        self._begin_op()
        if not servers:
            servers = '*'
        server_list = []
        # 
        server_pattern = ' '.join(servers)
        servers = [x.replace('*', '%') for x in servers]
        servers = [x.replace('?', '_') for x in servers]
        for s in servers:
            query = self.db_session.query(ServerSM)\
                    .filter(ServerSM.name.like(s))\
                    .order_by(ServerSM.name)
            if sg_name:
                if sg_name not in list_all_sgs(self.db_session):
                    raise NoSgFound(sg_name)
                query = query.join(ServerGroup,
                         ServerGroup.id_ == ServerSM.sg_id)\
                        .filter(ServerGroup.name == sg_name)
            server_list.extend(query.all())
        replica_sg = get_mastersm_replica_sg(self.db_session)
        if not show_all:
            replica_sg = get_mastersm_replica_sg(self.db_session)
            server_list = [s for s in server_list if s.sg != replica_sg ]
        if show_active:
            server_list = [s for s in server_list if (not s.is_disabled())]
        if not server_list:
            raise NoServerFound('*')
        server_list = [ s.to_engine_brief(time_format=self.time_format)
                            for s in server_list ]
        server_list = sorted(server_list, key=lambda s: s['server_name'])
        self._finish_op()
        return server_list

    def show_dms_status(self):
        """
        Show DMS system status
        """
        result = {}
        try:
            result['show_replica_sg'] = self.show_replica_sg()
        except NoReplicaSgFound:
            result['show_replica_sg'] = None
        result['show_mastersm'] = self.show_mastersm()
        try:
            result['list_server'] = self.list_server()
        except NoServerFound:
            result['list_server'] = None
        result['list_pending_events'] = self.list_pending_events()
        return result



#!usr/bin/env python3.2
#
# Copyright (c) Net24 Limited, Christchurch, New Zealand 2011-2012
#       and     Voyager Internet Ltd, New Zealand, 2012-2013
#
#    This file is part of py-magcode-core.
#
#    Py-magcode-core is free software: you can redistribute it and/or modify
#    it under the terms of the GNU  General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Py-magcode-core is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU  General Public License for more details.
#
#    You should have received a copy of the GNU  General Public License
#    along with py-magcode-core.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Module for Zone text manipuation utilities
"""

from io import StringIO
from textwrap import TextWrapper
import re

from pyparsing import ParseResults
import dns.name
import dns.ttl

from magcode.core.globals_ import *
from dms.globals_ import *
from dms.dns import RRTYPE_SOA
from dms.dns import RROP_DELETE
from dms.dns import is_inet_hostname
from dms.exceptions import NoPreviousLabelParseError
from dms.exceptions import Not7ValuesSOAParseError
from dms.exceptions import SOASerialMustBeInteger
from dms.exceptions import ZoneHasNoSOARecord
from dms.exceptions import BinaryFileError
from dms.exceptions import ZoneNameUndefined
from dms.exceptions import IncludeNotSupported
from dms.exceptions import GenerateNotSupported
from dms.exceptions import UpdateTypeNotSupported
from dms.exceptions import RropNotSupported
from dms.exceptions import TtlParseError
from dms.exceptions import HostnameParseError
from dms.exceptions import TtlInWrongPlace
from dms.exceptions import BadInitialZoneName
# Zone file parser is in a seperate module to contain symbol mess, and
# to aid in debugging from python3.2 command line.
from dms.zone_parser import zone_parser


rdata_re_null = re.compile(r'^""$|^\\#[ 	]+0$')

class DataToBind(object):
    """
    Objects of this class implement the transform from data to bind file
    output
    """
    
    def __init__(self, file=sys.stdout):
        self.comment_group_leader = settings['comment_group_leader']
        self.comment_rr_leader = settings['comment_rr_leader']
        self.comment_rrflags_leader = settings['comment_rrflags_leader']
        self.group_textwrapper = TextWrapper(
                            initial_indent=self.comment_group_leader + ' ',
                            subsequent_indent=self.comment_group_leader + ' ')
        self.rr_textwrapper = TextWrapper(
                            initial_indent=self.comment_rr_leader + ' ',
                            subsequent_indent=self.comment_rr_leader + ' ')
        self.rrflags_textwrapper = TextWrapper(
                            initial_indent=self.comment_rrflags_leader,
                            subsequent_indent=self.comment_rrflags_leader)
        self.for_bind = False
        self.file = file

    def print_group_comment(self, rr_group):
        if not rr_group.get('comment'):
            return
        comment = rr_group['comment']
        if (comment.find('\n') < 0):
            # If comment does not have any linefeeds, wrap it.
            print(self.group_textwrapper.fill(rr_group['comment']),
                    file=self.file)
            return
        # Print out
        comment = comment.split('\n')[:-1]
        for line in comment:
            print(self.comment_group_leader + ' ' + line, file=self.file)


    def print_rr_comment(self, rr):
        if not rr.get('comment'):
            return
        comment = rr['comment']
        if (comment.find('\n') < 0):
            # If comment does not have any linefeeds, wrap it.
            print(self.rr_textwrapper.fill(comment),
                    file=self.file)
            return
        # Print out
        comment = comment.split('\n')[:-1]
        for line in comment:
            print(self.comment_rr_leader + ' ' + line, file=self.file)

    def print_rrflags_comment(self, rr):
        if (not rr.get('lock_ptr') and not rr.get('disable')
                and not rr.get('reference') and not rr.get('track_reverse')):
            return
        rr_flags_strs = []
        if rr['lock_ptr']:
            rr_flags_strs.append(settings['rr_flag_lockptr'])
        if rr['disable']:
            rr_flags_strs.append(settings['rr_flag_disable'])
        if rr['track_reverse']:
            rr_flags_strs.append(settings['rr_flag_trackrev'])
        if rr['reference']:
            rr_flags_strs.append(settings['rr_flag_ref'] + rr['reference'])
        rr_flags_str = ' '.join(rr_flags_strs)
        print(self.rrflags_textwrapper.fill(rr_flags_str), file=self.file)

    def print_rr(self, rr, label, ttl):
        if (self.for_bind and rr.get('disable') != None and rr['disable']):
            label = ';' + label
        print ('%-15s %-7s %-7s %-15s %s' 
                % (label, ttl, rr['class'], rr['type'], rr['rdata']),
                file=self.file)
    
    def print_soa(self, rr, label, ttl):
        rdata = rr['rdata'].split()
        print ('%-15s %-7s %-7s %-15s ( %-12s ;Master NS' 
                % (label, ttl, rr['class'], rr['type'], rdata[0]),
                file=self.file)
        print ('%-47s %-12s ;RP email' % (' ', rdata[1]), file=self.file)
        print ('%-47s %-12s ;Serial yyyymmddnn' % (' ', rdata[2]),
                file=self.file)
        print ('%-47s %-12s ;Refresh' % (' ', rdata[3]), file=self.file)
        print ('%-47s %-12s ;Retry' % (' ', rdata[4]), file=self.file)
        print ('%-47s %-12s ;Expire' % (' ', rdata[5]), file=self.file)
        print ('%-47s %-12s ;Minimum/Ncache' % (' ', rdata[6]), file=self.file)
        print ('%-47s %-12s' % (' ', ')'), file=self.file)

    def print_rr_group(self, rr_group, sort_reverse=False, reference=None):
        """
        Print an rr_group

        This is a bit of a mess, but it gets the job done.
        """
        
        def sort_rr(rr):
            return(rr['label'], rr['type'], rr['rdata'])
       
        if rr_group.get('comment'):
            self.print_group_comment(rr_group)
        previous_label = ''
        for rr in sorted(rr_group['rrs'], key=sort_rr, reverse=sort_reverse):
            self.print_rr_comment(rr)
            if reference and rr['type'] == RRTYPE_SOA:
                rr['reference'] = reference
            self.print_rrflags_comment(rr)
            ttl = rr['ttl'] if rr.get('ttl') else '    '
            label = rr['label'] if (rr.get('label') != previous_label) \
                                    else '    '
            if rr['type'] == RRTYPE_SOA:
                self.print_soa(rr, label, ttl)
            else:
                self.print_rr(rr, label, ttl)
            previous_label = rr['label']
        print('\n\n', file=self.file, end='')

    def __call__(self, zi_data, name=None, reference=None, for_bind=False,
            file=None, no_info_header=False):
        """
        Construct a bind file as a multi-line string, from
        zi_data
        """
        # if zi_data is blank, get out of here...
        if not zi_data:
            return ''
        # Do file/IO house keeping first
        if not file:
            self.file = StringIO()
            return_string = True
        else:
            self.file = file
            return_string = False
        
        # Save bind_p
        self.for_bind = for_bind

        # Set $TTL and $ORIGIN if given
        print ('$TTL %s' % zi_data['zone_ttl'], file=self.file)
        if name:
            print ('$ORIGIN %s' % name, file=self.file)
        print(file=self.file)

        # Add reference comment if reference given
        zi_id = zi_data.get('zi_id')
        zi_change_by = zi_data.get('change_by')
        zi_ctime = zi_data.get('ctime')
        zi_mtime = zi_data.get('mtime')
        zi_ptime = zi_data.get('ptime')
        if not no_info_header and (reference or name or zi_id):
            # Trailing double line feed for readability
            out = ";\n"
            if name:
                out += "; Zone:       %s\n" % name
            if reference:
                out += "; Reference:  %s\n" % reference
            if zi_change_by:
                out += "; change_by:  %s\n" % zi_change_by
            if zi_id:
                out += "; zi_id:      %s\n" % zi_id
            if zi_ctime:
                out += "; zi_ctime:   %s\n" % zi_ctime
            if zi_mtime:
                out += "; zi_mtime:   %s\n" % zi_mtime
            if zi_ptime:
                out += "; zi_ptime:   %s\n" % zi_ptime
            out += ";\n\n"
            print(out, file=self.file)

        # Index rr_groups, for printing
        rr_groups = {}
        flimflam_gid = '' 
        for rr_group in zi_data['rr_groups']:
            group_tag = rr_group.get('tag')
            group_comment = rr_group.get('comment')
            if group_tag == settings['apex_rr_tag']:
                group_id = group_tag
            elif group_comment:
                group_id = group_comment
            elif group_tag:
                group_id = group_tag
            else:
                group_id = str(flimflam_gid)
                if flimflam_gid == '':
                    flimflam_gid = 0
                flimflam_gid += 1
            rr_groups[group_id] = rr_group
        
        # Print Apex Records if there are any
        rr_group = rr_groups.get(settings['apex_rr_tag'])
        if rr_group:
            self.print_rr_group(rr_group, sort_reverse=True, 
                    reference=reference)
            del rr_groups[settings['apex_rr_tag']]

        # Print the rest, followed by default group
        default_group = rr_groups.pop('', None)
        for rr_group in sorted(rr_groups):
            self.print_rr_group(rr_groups[rr_group])
        if default_group:
            self.print_rr_group(default_group)

        # clean up
        if return_string:
            result = self.file.getvalue()
            self.file.close()
            return result

def data_to_bind(zi_data, name=None, file=None, 
        for_bind=False, reference=None, no_info_header=False):
    """
    Translate data to bind file output
    """
    transform = DataToBind()
    return transform(zi_data, name=name, file=file, 
                for_bind=for_bind, reference=reference, 
                no_info_header=no_info_header)

def _validate_pyparsing_hostname(name, data, text):
    try:
        thing = dns.name.from_text(text)
    except Exception as exc:
        raise HostnameParseError(name, data, text, str(exc))
    if not is_inet_hostname(text):
        raise HostnameParseError(name, data, text, None)
    if not text.endswith('.'):
        raise HostnameParseError(name, data, text, "must end with '.'.")

def _validate_pyparsing_ttl(name, data, text):
    """
    Validate a ttl value
    """
    if len(text) > 20:
        raise TtlParseError(name, data, text, "longer than 20 chars.")
    try:
        thing = dns.ttl.from_text(text)
    except Exception as exc:
        raise TtlParseError(name, data, text, str(exc))

def bind_to_data(bind_file, name=None, use_origin_as_name=False, 
                    update_mode=False):
    """
    Construct zi_data, taking a bind file as input.  Can be a string, 
    or file handle.
    """
    def validate_initial_name():
        if not name:
            return
        if isinstance(bind_file, StringIO):
            input_thing = 'StringIO object'
        elif file_name:
            input_thing = file_name
        else:
            input_thing = ('FD %s' % str(bind_file.fileno()))
        try:
            thing = dns.name.from_text(name)
        except Exception as exc:
            raise BadInitialZoneName(input_thing, name, str(exc))
        if not is_inet_hostname(name):
            raise BadInitialZoneName(input_thing, name, None)
        if not name.endswith('.'):
            raise BadInitialZoneName(input_thing, name, "must end with '.'.")

    def check_name_defined():
        if not name:
            if isinstance(bind_file, StringIO):
                raise ZoneNameUndefined('StringIO object')
            elif file_name:
                raise ZoneNameUndefined(file_name)
            else:
                raise ZoneNameUndefined('FD %s' % str(bind_file.fileno()))

    file_name = None
    if isinstance(bind_file, str):
        file_name = bind_file
        # Open file
        # Check that file is not 'binary'
        bfile = open(file_name, mode='rb')
        bs = bfile.read(256)
        bfile.close()
        try:
            if not bs.decode().replace('\n', ' ')\
                    .replace('\t', ' ').isprintable():
                raise BinaryFileError(bind_file) 
        except UnicodeError:
                raise BinaryFileError(bind_file) 

        bind_file = open(file_name, mode='rt')

    # Check Initial name and if it is garbage do something appropriate
    try:
        validate_initial_name()
    except BadInitialZoneName as exc:
        if not use_origin_as_name:
            if file_name:
                bind_file.close()
            raise exc
        name = None

    # Feed through pyparsing to get back a parse result we can traverse
    # Error Exceptions handled at higher level for error processing
    try:
        zone_parse = zone_parser.parseFile(bind_file, parseAll=True)
    finally:
        if file_name:
            bind_file.close()

    # Turn parse result into a JSON zi data structure
    zi_data = {'rr_groups':[], }
    # List of all rrs for sorting out Apex and SOA data at end
    # of loop
    rr_list = []
    # Loop data variables
    in_rr_group = False
    comment_rr = None
    comment_rrflags = None
    update_type = None
    zone_reference = None
    comment_group = None
    previous_label = None
    origin = name if name else None
    ttl_seen = False
    ttl = None
    in_rr_prologue = True
    rr_group = {'rrs':[]}
    for thing in zone_parse:
        if (isinstance(thing, dict)
                and thing['type'] == '$ORIGIN'):
            _validate_pyparsing_hostname(name, thing, thing['origin'])
            origin = thing['origin']
            continue

        if (isinstance(thing, dict)
                and thing['type'] == '$TTL'):
            if ttl_seen:
                raise TtlInWrongPlace(name, thing, file_name)
            _validate_pyparsing_ttl(name, thing, thing['ttl'])
            ttl_seen = True
            zi_data['zone_ttl'] = thing['ttl']
            continue

        if (isinstance(thing, dict)
                and thing['type'] == '$INCLUDE'):
            raise IncludeNotSupported(name, thing)

        if (isinstance(thing, dict)
                and thing['type'] == '$GENERATE'):
            raise GenerateNotSupported(name, thing)

        if (isinstance(thing, dict)
                and thing['type'] == '$UPDATE_TYPE'):
            if not update_mode:
                raise UpdateTypeNotSupported(name, thing)
            update_type = thing['update_type']

        if (isinstance(thing, dict) 
                and thing['type'] == 'comment_rr'):
            #Process an  RR comment
            comment_rr = thing
            comment_rr.pop('comment_type', None)
            continue
        
        if (isinstance(thing, dict) 
                and thing['type'] == 'comment_rrflags'):
            #Process rr_flags
            comment_rrflags = thing
            comment_rrflags.pop('comment_type', None)
            continue

        if (isinstance(thing, dict)
                and thing['type'] == 'comment_group'):
            #Process a group comment
            comment_group = thing
            comment_group.pop('comment_type', None)
            if not in_rr_group:
                # Start New RR Group from previous blank lines
                in_rr_group = True
                rr_group.update(comment_group)
            else:
                # Start new RR_Group
                zi_data['rr_groups'].append(rr_group)
                rr_group = {'rrs':[]}
                rr_group.update(comment_group)
            continue

        if isinstance(thing, ParseResults):
            # $TTL should have happened by now
            ttl_seen = True
            if in_rr_prologue:
                in_rr_prologue = False
                # if no name, should have seen $ORIGIN by now
                if origin and use_origin_as_name:
                    name = origin
                check_name_defined()

            # Start process RRs
            if not in_rr_group:
                in_rr_group = True
            # Process an RR
            rr = {}
            # Have to break down to keys we accept  - security
            # Sort out RR label - if none, use last seen value of label
            rr['label'] = thing.get('label')
            if not rr['label']:
                if not previous_label:
                    # This should not happen!
                    # This error is feature specific to the parser design,
                    # and should be raised here
                    raise NoPreviousLabelParseError(domain=name)
                rr['label'] = previous_label
            else:
                previous_label = rr['label']

            # Apply $ORIGIN to label.  This will be relativized when 
            # actual RR object is created.
            if origin and not rr['label'].endswith('.'):
                if rr['label'] == '@':
                    rr['label'] = origin
                else:
                    rr['label'] = '.'.join((rr['label'],origin))

            # Add preceding rr_flags and comment_rr to RR
            if comment_rr:
                rr.update(comment_rr)
                comment_rr = None

            # Decode rr_flags
            rr['lock_ptr'] = False
            rr['disable'] = False 
            rr['force_reverse'] = False 
            rr['track_reverse'] = False 
            rr['reference'] = None 
            rr['update_op'] = None
            if comment_rrflags:
                rr_flags = comment_rrflags['rr_flags'].strip()
                if rr_flags.find(settings['rr_flag_forcerev']) >= 0:
                    rr['force_reverse'] = True 
                if rr_flags.find(settings['rr_flag_trackrev']) >= 0:
                    rr['track_reverse'] = True 
                if rr_flags.find(settings['rr_flag_lockptr']) >= 0 :
                    rr['lock_ptr'] = True 
                if rr_flags.find(settings['rr_flag_disable']) >= 0 :
                    rr['disable'] = True 
                rr_flags = rr_flags.split()
                for rr_flag in rr_flags:
                    if not (rr_flag.find(settings['rr_flag_ref']) >= 0):
                        continue
                    reference = rr_flag[len(settings['rr_flag_ref']):]
                    if (thing.get('type') and thing['type'] == RRTYPE_SOA
                            and not rr['disable']):
                        zone_reference = reference
                        break
                    rr['reference'] = reference
                    break
                for rr_flag in rr_flags:
                    if not (rr_flag.find(settings['rr_flag_rrop']) >= 0):
                        continue
                    update_op = rr_flag[len(settings['rr_flag_rrop']):]
                    if not update_mode:
                        raise RropNotSupported(name, comment_rrflags)
                    rr['update_op'] = update_op
                    break
                comment_rrflags = None
            
            # Unpack and decode rdata from pyparsing
            # This gives us the file location and output line showing position
            # for any rdata exceptions we throw later in 
            # dms.database.resource_record.data_to_rr()
            if thing.get('rdata'):
                rdata = thing['rdata']
                if isinstance(rdata, dict):
                    rr['rdata'] = rdata.get('rdata')
                    rr['rdata_pyparsing'] = rdata.get('pyparsing')
                else:
                    rr[rdata] = rdata
                if rr['update_op'] == RROP_DELETE:
                    # For delete update_op, transliterate rdata strings
                    if rdata_re_null.search(rr['rdata']):
                        rr['rdata'] = None
                        

            # Do type, class, and ttl
            for key in ('type', 'class', 'ttl'):
                if thing.get(key):
                    rr[key] = thing[key]
            # Add rr to rr_group and rr_list
            rr_group['rrs'].append(rr)
            rr_list.append(rr) 
            continue

        if thing == '\n':
            # Process a blank line
            if in_rr_group:
                in_rr_group = False
                # Add rr_group to list of rr_groups
                zi_data['rr_groups'].append(rr_group)
                rr_group = {'rrs':[]}
            continue

        # We don't care bout this 'thing'
        continue
    else:
        # clean up - if this is not done, not ending in a blank line will
        # lose records....
        if in_rr_group:
             # Add rr_group to list of rr_groups
             zi_data['rr_groups'].append(rr_group)
        
    # Fill in zi_data fields from SOA. Use first SOA found.  Will check
    # for duplicate SOA further in, as that may be recieved from DMI/DMS
    rr_soas = [rr for rr in rr_list if rr['type'] == RRTYPE_SOA]
    if not rr_soas:
        # Leave early as this might just be a zone file being loaded for
        # a use_apex_ns zone, in which case this does not matter
        check_name_defined()
        return (zi_data, name, update_type, zone_reference)
    rr_soa = rr_soas.pop(0)
    # Determine zone name if use_origin_as_name is set, by looking at SOA
    # record label
    if use_origin_as_name:
        if rr_soa['label'][-1] == '.':
            name = rr_soa['label']

    check_name_defined()
    
    zi_data['soa_ttl'] = rr_soa.get('ttl')
    # Parse SOA Rdata
    soa_rdata = rr_soa['rdata'].split()
    num_values = len(soa_rdata)
    if (num_values != 7):
        raise Not7ValuesSOAParseError(name,
                rr_soa)
    error_info = ''
    try:
        zi_data['soa_serial'] = int(soa_rdata[2])
    except ValueError as exc:
        error_info = str(exc)
    if error_info:
        raise SOASerialMustBeInteger(name,
                rr_soa)
    zi_data['soa_mname'] = soa_rdata[0]
    zi_data['soa_rname'] = soa_rdata[1]
    zi_data['soa_refresh'] = soa_rdata[3]
    zi_data['soa_retry'] = soa_rdata[4]
    zi_data['soa_expire'] = soa_rdata[5]
    zi_data['soa_minimum'] = soa_rdata[6]

    return (zi_data, name, update_type, zone_reference)











#!/usr/bin/env python3.2
#
# Copyright (c) Net24 Limited, Christchurch, New Zealand 2011-2012
#       and     Voyager Internet Ltd, New Zealand, 2012-2013
#
#    This file is part of py-magcode-core.
#
#    Py-magcode-core is free software: you can redistribute it and/or modify
#    it under the terms of the GNU  General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Py-magcode-core is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU  General Public License for more details.
#
#    You should have received a copy of the GNU  General Public License
#    along with py-magcode-core.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Server State Machines
"""


import socket
import errno
from random import random
from datetime import timedelta
from subprocess import check_call
from subprocess import check_output
from subprocess import CalledProcessError
from subprocess import STDOUT

import dns.name
import dns.rdatatype
import dns.rdataclass
import dns.message
import dns.exception
import dns.query
import dns.flags
import dns.rrset
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.exc import IntegrityError

from magcode.core.database import *
from magcode.core.database.state_machine import StateMachine
from magcode.core.database.state_machine import smregister
from magcode.core.database.state_machine import SMEvent
from magcode.core.database.state_machine import SMSyncEvent
from magcode.core.database.state_machine import StateMachineError
from magcode.core.database.state_machine import StateMachineFatalError
from magcode.core.database.event import eventregister
from magcode.core.database.event import synceventregister
from magcode.core.database.event import find_events
from magcode.core.database.event import ESTATE_SUCCESS
from magcode.core.database.event import create_event
from dms.database import zone_cfg
from dms.database.sg_utility import list_all_sgs
from dms.database.sg_utility import find_sg_byname
from dms.exceptions import ServerExists
from dms.exceptions import NoServerFound
from dms.exceptions import NoServerFoundByAddress
from dms.exceptions import ServerAddressExists
from dms.exceptions import ServerNotDisabled
from dms.exceptions import NoSgFound
from dms.exceptions import ServerSmFailure
from dms.dns import RRTYPE_SOA


# Some constants
SSTATE_CONFIG = "CONFIG"
SSTATE_OK = "OK"
SSTATE_RETRY = "RETRY"
SSTATE_BROKEN = "BROKEN"
SSTATE_DISABLED = "DISABLED"

SSTYPE_BIND9 = 'bind9'
SSTYPE_NSD3 = 'nsd3'
server_types = [SSTYPE_BIND9, SSTYPE_NSD3]

class CantRsyncServer(StateMachineError):
    """
    Rsync to the server failed.
    """
class CantRsyncDnssecKeysDrServer(CantRsyncServer):
    """
    Rsync of DNSSEC keys to DR replica server failed
    """
class CantRndcServer(StateMachineError):
    """
    Rndc to the server failed.
    """
class CantSoaQueryServer(StateMachineError):
    """
    Cant SOA query Server - retry
    """
class BrokenServer(CantSoaQueryServer):
    """
    Broken action - DNS server non functional. Keep retrying.
    """
class ServerAlreadyDisabled(StateMachineFatalError):
    """
    Server already disabled.
    """
class ServerAlreadyEnabled(StateMachineFatalError):
    """
    Server already enabled.
    """
class ServerEnableFailure(StateMachineFatalError):
    """
    Server already enabled.
    """

# Server State Machine Events
@synceventregister
class ServerSMEnable(SMSyncEvent):
    """
    Enable Server
    """
    pass

@synceventregister
class ServerSMDisable(SMSyncEvent):
    """
    Disable Server
    """
@synceventregister
class ServerSMReset(SMSyncEvent):
    """
    Reset Server
    """

@eventregister
class ServerSMConfigure(SMEvent):
    """
    Server configured
    """
    pass

@eventregister
class ServerSMConfigChange(SMEvent):
    """
    Configuration change for configured server
    """

@eventregister
class ServerSMCheckServer(SMEvent):
    """
    Check a configured server to check that it is running
    """

@smregister
class ServerSM(StateMachine):
    """
    Server State Machine

    Implements Server State machine
    """
    _table = 'sm_servers'
    _sm_events = (ServerSMEnable, ServerSMDisable, 
                ServerSMReset, ServerSMConfigure,
                ServerSMConfigChange, ServerSMCheckServer )

    def __init__(self, server_name, address, sg_name, server_type,
                    ssh_address):
        """
        Create a server SM object
        """
        self.name = server_name
        self.address = address
        self.server_type = server_type
        self.ssh_address = ssh_address
        self.state = SSTATE_DISABLED
        self.ctime = None
        self.mtime = None
        self.last_reply = None
        self.retry_msg = None
    
    def set_sg(self, sg):
        """
        Set the server group this zone is served on.
        """
        if hasattr(self, 'sg') and self.sg:
            old_sg = self.sg
            old_sg.servers.remove(self)
            # do the following if relationships are set up for lazy='dynamic'
            #self.sg = None
            #del(self.sg)
        sg.servers.append(self)
        self.sg = sg

    def _to_engine_stuff(self, time_format):
        """
        Backend common function to fill out timestamps etc for
        to_engine methods.
        """
        if hasattr(self, 'sg') and self.sg:
            sg_name = self.sg.name
        else:
            sg_name = None
        if not time_format:
            ctime = (self.ctime.isoformat(sep=' ') 
                               if self.ctime else None)
            mtime = (self.mtime.isoformat(sep=' ') 
                               if self.mtime else None)
            last_reply = (self.last_reply.isoformat(sep=' ') 
                               if self.last_reply else None)
        else:
            ctime = (self.ctime.strftime(time_format) 
                               if self.ctime else None)
            mtime = (self.mtime.strftime(time_format) 
                               if self.mtime else None)
            last_reply = (self.last_reply.strftime(time_format) 
                               if self.last_reply else None)
        return (sg_name, ctime, mtime, last_reply)


    def to_engine_brief(self, time_format=None):
        """
        Output server SM attributes as JSON
        """
        sg_name, ctime, mtime, last_reply = self._to_engine_stuff(time_format)
        return {'server_id': self.id_, 'server_name': self.name,
                'address': self.address, 'state': self.state,
                'ctime': ctime, 'mtime': mtime,
                'is_master': self.is_master(),
                'last_reply': last_reply,
                'retry_msg': self.retry_msg,
                'ssh_address': self.ssh_address}

    def to_engine(self, time_format=None):
        """
        Output server SM attributes as JSON
        """
        sg_name, ctime, mtime, last_reply = self._to_engine_stuff(time_format)
        return {'server_id': self.id_, 'server_name': self.name,
                'address': self.address,
                'state': self.state, 'server_type': self.server_type, 
                'sg_id': self.sg_id, 'sg_name': sg_name,
                'ctime': ctime, 'mtime': mtime,
                'is_master': self.is_master(),
                'last_reply': last_reply,
                'retry_msg': self.retry_msg,
                'ssh_address': self.ssh_address, 
                'zone_count': self.zone_count}

    def is_this_server(self):
        """
        Return whether this server is actually this server

        Server can be congfigured in DB for DR use
        """
        return self.address in settings['this_servers_addresses']

    def is_disabled(self):
        """
        Return whether the server is DISABLED or not
        """
        return self.state == SSTATE_DISABLED

    def query_is_disabled(self, query):
        """
        Add DISABLED query term 
        """
        return query.filter(ServerSM.state == SSTATE_DISABLED)

    def query_is_not_disabled(self, query):
        """
        Add not DISABLED query term 
        """
        return query.filter(ServerSM.state != SSTATE_DISABLED)

    def is_master(self):
        """
        Return if this server definition is currently the master
        """
        if (hasattr(self, 'master_sm') and self.master_sm):
            return True
        return False

    def _rsync_dnssec_keys(self, event):
        """
        Rsync DNSSEC key files across to DR replica server
        """
        if not(self.sg and hasattr(self.sg, 'master_sm') 
                and self.sg.master_sm):
            # OK, we are not interested in DNSSEC stuff
            return
        key_dir = settings['master_dnssec_key_dir'] 
        try:
            cmdline = (settings['rsync_path'] 
                    + ' ' + settings['rsync_dnssec_args'] 
                    + ' --password-file ' 
                    + settings['rsync_dnssec_password_file'] 
                    + ' ' + key_dir + '/'
                    + ' ' + settings['rsync_dnssec_target'])
            # Add IPv6 address squares
            address_string = '[' + self.address + ']' \
                        if self.address.find(':') else self.address
            cmdline_str = cmdline % address_string
            cmdline = cmdline_str.split(' ')
            output = check_output(cmdline, stderr=STDOUT)
        except CalledProcessError as exc:
            if exc.output:
                # Here is something a bit untidy
                output = str(exc.output)[2:-3].replace('\\n', ', ')
                msg = (
                    "Server '%s': failed to rsync dnssec keys, %s, %s" 
                    % (self.name, str(exc), output))
            else:
                msg = (
                    "Server '%s': failed to rsync dnssec keys, %s" 
                    % (self.name, str(exc)))
            raise CantRsyncDnssecKeysDrServer(msg)

    def _rsync_includes(self, event):
        """
        Rsync include files across to a server
        """
        include_dir = self.sg.get_include_dir()
        try:
            cmdline = (settings['rsync_path'] + ' ' + settings['rsync_args'] 
                    + ' --password-file ' + settings['rsync_password_file'] 
                    + ' ' + include_dir + '/' + ' ' + settings['rsync_target'])
            # Add IPv6 address squares
            address_string = '[' + self.address + ']' \
                        if self.address.find(':') else self.address
            cmdline_str = cmdline % address_string
            cmdline = cmdline_str.split(' ')
            output = check_output(cmdline, stderr=STDOUT)
        except CalledProcessError as exc:
            if exc.output:
                # Here is something a bit untidy
                output = str(exc.output)[2:-3].replace('\\n', ', ')
                msg = (
                    "Server '%s': failed to rsync include files, %s, %s" 
                    % (self.name, str(exc), output))
            else:
                msg = (
                    "Server '%s': failed to rsync include files, %s" 
                    % (self.name, str(exc)))
            raise CantRsyncServer(msg)

    def _rndc_server(self, event, *rndc_args):
        """
        Run rndc
        """
        output = ''
        try:
            cmdline = [settings['rndc_path']]
            if not self.is_master():
                cmdline.extend([ '-s', self.name])
            cmdline.extend(rndc_args)
            output = check_output(cmdline, universal_newlines=True)
        except CalledProcessError as exc:
            msg = ("Server '%s': %s failed, %s" 
                    % (self.name, settings['rndc_path'], str(exc)))
            raise CantRndcServer(msg)
        if (rndc_args[-1] == 'status' and output):
            try:
                output = output.split('\n')
                output = [s for s in output
                        if s.find(settings['bind9_zone_count_tag']) != -1]
                if not len(output):
                    raise IndexError("Can't gather zone Count")
                zone_stuff = output[0].split()
                self.zone_count = int(zone_stuff[-1])
            except (IndexError, ValueError)as exc:
                log_error("Server '%s': can't gather zone count")

    def _soa_query_server(self, zone_name):
        """
        Use dnspython to read the SOA record of a Zone from the DNS server.

        This function returns whether the server is speaking inteligible DNS
        or not.  It function is as a keep alive check.
        """
        zone = dns.name.from_text(zone_name)
        rdtype = dns.rdatatype.from_text(RRTYPE_SOA)
        rdclass = dns.rdataclass.IN
        query = dns.message.make_query(zone, rdtype, rdclass)
        exc = None
        try:
            # Use TCP as dnspython can't track replies to multithreaded
            # queries
            answer = dns.query.tcp(query, self.address,
                    timeout=float(settings['dns_query_timeout']))
            if not query.is_response(answer):
                msg = ("Server '%s': SOA query - reply from unexpected source,"
                        " retrying" % self.name)
                raise CantSoaQueryServer(msg)
        except dns.query.BadResponse as exc:
            msg = ("Server '%s': SOA query - received incorrectly"
                    " formatted query."  % self.name)
            raise BrokenServer(msg)
        except dns.exception.Timeout:
            msg =  ("Server '%s': SOA query - timeout waiting for response,"
                    " retrying" % self.name)
            raise CantSoaQueryServer(msg)
        except dns.query.UnexpectedSource as exc:
            # For UDP, FormError and BadResponse here are also failures
            msg = ("Server '%s': SOA query - reply from unexpected source,"
                    " retrying" % self.name)
            raise CantSoaQueryServer(msg)
        except dns.exception.FormError as exc:
            msg = ("Server '%s': SOA query - remote responded incorrectly"
                    " formatted query."  % self.name)
            raise BrokenServer(msg)
        except (socket.error, OSError, IOError) as exc:
            if errno in (errno.EACCES, errno.EPERM, errno.ECONNREFUSED, 
                    errno.ENETUNREACH, errno.ETIMEDOUT):
                msg = ("Server '%s': SOA query - can't query server %s - %s"
                        % (self.name, self.address, exc.strerror))
                raise CantSoaQueryServer(msg)
            msg = ("Server '%s': server %s, SOA query - fatal error %s."
                   % (self.name, self.address, exc.strerror))
            raise BrokenServer(msg)
        finally:
            # Clean up memory
            del query
        try:
            # Check and process result codes
            # with 0, check that answer.answer contains stuff, and check type of
            # 1st element is dns.rrset.RRset via isinstance()
            rcode = answer.rcode()
            rcode_text = dns.rcode.to_text(answer.rcode())
            if rcode in _soaquery_rcodes['success']:
                if (len(answer.answer) 
                        and isinstance(answer.answer[0], dns.rrset.RRset)):
                    return
                msg = ("Server '%s': SOA query - bad response received."
                        % self.name)
                raise BrokenServer(msg)
            elif rcode in _soaquery_rcodes['ok']:
                return
            elif rcode in _soaquery_rcodes['retry']:
                msg = ("Server '%s': SOA query - temporary failure - rcode '%s'"
                        % (self.name, rcode_text))
                raise CantSoaQueryServer(msg)
            elif rcode in _soaquery_rcodes['broken']:
                msg = ("Server '%s': SOA query - broken - rcode '%s'"
                        % (self.name, rcode_text))
                raise BrokenServer(msg)
            else:
                msg = ("Server '%s': SOA query - response with indeterminate"
                        " error - broken?" % self.name)
                raise BrokenServer(msg)
                
        finally:
            # clean up memory
            del answer

    def _process_sm_exc(self, db_session, exc, msg, new_state=None):
        """
        Process a state machine exception.  For putting as much code as
        possible on a common call path.
        """
        delay_factor = 1 + random()
        delay_period = timedelta(
                minutes=delay_factor*float(settings['master_hold_timeout']))
        old_state = self.state
        if new_state:
            self.state = new_state
        log_msg = str(exc)
        log_info(log_msg)
        self.retry_msg = log_msg
        create_event(ServerSMConfigure, db_session=db_session,
            sm_id=self.id_, server_id=self.id_, delay=delay_period, 
            server_name=self.name)
        state_str = ''
        if old_state != self.state:
            state_str = ("old state %s, new state %s - " 
                                    % (old_state, self.state) )
        return (RCODE_OK, "Server '%s': %s%s" 
                    % (self.name, state_str, msg))
    
    def _create_check(self, event):
        """
        Process a state machine exception.  For putting as much code as
        possible on a common call path.
        """
        db_session = event.db_session

        # Check every half to full holdout time
        master_hold_timeout = float(settings['master_hold_timeout'])
        delay_factor = (1 + random()) * 0.5
        delay_period = timedelta(minutes=delay_factor*master_hold_timeout)
        hold_period = timedelta(minutes=master_hold_timeout)

        # See if a check event already exists for this ServerSM
        current_checks = find_events(ServerSMCheckServer, db_session,
                                                    server_id=self.id_)
        current_checks = [e for e in current_checks if e.id_ != event.id_]
        if len(current_checks):
            return

        create_event(ServerSMCheckServer, db_session=db_session,
            sm_id=self.id_, server_id=self.id_, delay=delay_period, 
            server_name=self.name)

    def _disable(self, event):
        """
        Disable the Server
        """
        self.state = SSTATE_DISABLED
        self.last_reply = None
        self.retry_msg = None
        return (RCODE_OK, "Server '%s': disabled" % self.name)

    def _already_disabled(self, event):
        """
        Disable the Server
        """
        raise ServerAlreadyDisabled("server already disabled")

    def _enable(self, event):
        """
        Enable the Server
        """
        try:
            query = event.db_session.query(ServerSM)\
                    .filter(ServerSM.address == self.address)\
                    .filter(ServerSM.state != SSTATE_DISABLED)
            result = query.all()
            if result:
                raise ServerEnableFailure(
                    "Server '%s' - server '%s' with same address enabled"
                        % (self.name, result[0].name))
        except NoResultFound:
            pass
        self.state = SSTATE_CONFIG
        create_event(ServerSMConfigure, db_session=event.db_session,
                sm_id=self.id_, server_id=self.id_, server_name=self.name)
        return (RCODE_OK, "Server '%s': enabling" % self.name)

    def _already_enabled(self, event):
        """
        Disable the Server
        """
        raise ServerAlreadyEnabled("server already enabled")

    def _config_change(self, event):
        """
        Start the reconfiguration process
        """
        self.state = SSTATE_CONFIG
        self.retry_msg = None
        create_event(ServerSMConfigure, db_session=event.db_session,
                sm_id=self.id_, server_id=self.id_, server_name=self.name)
        return (RCODE_OK, "Server '%s': reconfiguring" % self.name)

    def _configure(self, event):
        """
        Configure a server
        """
        db_session = event.db_session
        if not self.is_this_server():
            # rsync configuration
            try:
                self._rsync_includes(event)
                self._rsync_dnssec_keys(event)
            except CantRsyncServer as exc:
                return self._process_sm_exc(db_session, exc,
                        "retrying config process", SSTATE_RETRY)

        # Test that server is talking sanely
        try:
            self._soa_query_server(settings['serversm_soaquery_domain'])
        except BrokenServer as exc:
            return self._process_sm_exc(db_session, exc, 
                        "retrying config process", SSTATE_BROKEN)
        except CantSoaQueryServer as exc:
            return self._process_sm_exc(db_session, exc, 
                        "retrying config process", SSTATE_RETRY)

        if self.is_this_server():
            # This means we are master - DON'T DO ANYTHING as this can
            # cause an rndc race in bind, which can trash dynamic zones.
            self.state = SSTATE_OK
            self.retry_msg = None
            self.last_reply = db_clock_time(db_session)
            self._create_check(event)
            return (RCODE_OK, 
                    "Server '%s': now master - SSM slot reserved" 
                        % self.name)
        if self.server_type == 'bind9':
            self.last_reply = db_clock_time(db_session)
            db_session.flush()
            try:
                self._rndc_server(event, 'reconfig')
            except CantRndcServer as exc:
                return self._process_sm_exc(db_session, 
                        exc, "retrying config process", SSTATE_RETRY)
        
        self.state = SSTATE_OK
        self.retry_msg = None
        # create check event
        self.last_reply = db_clock_time(db_session)
        self._create_check(event)
        return (RCODE_OK, "Server '%s': configured" % self.name)

    def _check_server(self, event):
        """
        Check that a server is running and gather some statistics
        """
        db_session = event.db_session

        # Test that server is talking sanely
        try:
            self._soa_query_server(settings['serversm_soaquery_domain'])
        except BrokenServer as exc:
            return self._process_sm_exc(db_session, exc, 
                        "retrying config process", SSTATE_BROKEN)
        except CantSoaQueryServer as exc:
            return self._process_sm_exc(db_session, exc, 
                        "retrying config process", SSTATE_RETRY)

        # Check that rndc is working and gather some stats
        if self.server_type == 'bind9':
            self.last_reply = db_clock_time(db_session)
            db_session.flush()
            try:
                self._rndc_server(event, 'status')
            except CantRndcServer as exc:
                return self._process_sm_exc(db_session, 
                        exc, "retrying config process", SSTATE_RETRY)

        self.last_reply = db_clock_time(db_session)
        self._create_check(event)
        return (RCODE_OK, "Server '%s': alls well" % self.name)

    _sm_table = {   
            SSTATE_DISABLED: {
                ServerSMEnable: _enable,
                ServerSMDisable: _already_disabled,
                },
            SSTATE_CONFIG: {
                ServerSMDisable: _disable,
                ServerSMEnable: _already_enabled,
                ServerSMConfigure: _configure,
                ServerSMReset: _config_change,
                },
            SSTATE_RETRY: {
                ServerSMDisable: _disable,
                ServerSMEnable: _already_enabled,
                ServerSMConfigure: _configure,
                ServerSMReset: _config_change,
                },
            SSTATE_BROKEN: {
                ServerSMDisable: _disable,
                ServerSMEnable: _already_enabled,
                ServerSMConfigure: _configure,
                ServerSMReset: _config_change,
                },
            SSTATE_OK: {
                ServerSMDisable: _disable,
                ServerSMEnable: _already_enabled,
                ServerSMConfigChange: _config_change,
                ServerSMReset: _config_change,
                ServerSMCheckServer: _check_server,
                },
            }
                

def exec_server_sm(server_sm, sync_event_type, 
                exception_type=ServerSmFailure,
                **event_kwargs):
    """
    Execute a synchronous event of the server state machine
    """
    if not issubclass(sync_event_type, SMSyncEvent):
        raise TypeError("'%s' is not a Synchonous Event." % sync_event_type)

    event = sync_event_type(sm_id=server_sm.id_,
                                    server_id=server_sm.id_,
                                    **event_kwargs)
    results = event.execute()
    if results['state'] != ESTATE_SUCCESS:
        # By std Python convention exceptions don't have default value
        # arguments. Do the following to take care of 2 or 3 argument
        # variants for the exception.
        raise exception_type(server_sm.name, results['message'], results)
    return results

def new_server(db_session, server_name, address, sg_name, server_type=None,
                ssh_address=None):
    """
    Create a new server
    """
    server_name = server_name.lower()
    if server_name.endswith('.'):
        server_name = server_name[:-1]
    if not sg_name:
        sg_name = zone_cfg.get_row_exc(db_session, 'default_sg')
    if not sg_name in list_all_sgs(db_session):
        raise NoSgFound(sg_name)
    try:
        server_list = db_session.query(ServerSM)\
                        .filter(ServerSM.name == server_name).all()
        if len(server_list):
            raise ServerExists(server_name)
    except NoResultFound:
        pass
    if not server_type:
        server_type = zone_cfg.get_row(db_session, 'default_stype', 
                        raise_exc=True)
    server_sm = ServerSM(server_name, address, sg_name, server_type,
                            ssh_address)
    try:
        db_session.add(server_sm)
        db_session.flush()
    except IntegrityError as exc:
        raise ServerAddressExists(address)
    sg = find_sg_byname(db_session, sg_name, raise_exc=True)
    server_sm.set_sg(sg)
    db_session.flush()
    return server_sm


def del_server(db_session, server_name):
    """
    Delete a server
    """
    # Get the Server from the DB.
    try:
        server_sm = db_session.query(ServerSM)\
                    .filter(ServerSM.name == server_name).one()
    except NoResultFound:
        raise NoServerFound(server_name)
    if not server_sm.is_disabled():
        raise ServerNotDisabled(server_name)
    db_session.delete(server_sm)
    db_session.flush()

def find_server_byname(db_session, server_name, raise_exc=True):
    """
    Find a server by name
    """
    query = db_session.query(ServerSM)\
            .filter(ServerSM.name == server_name)
    try:
        server_sm = query.one()
    except NoResultFound:
        server_sm = None
    if raise_exc and not server_sm:
        raise NoServerFound(server_name)
    return server_sm

def find_server_byaddress(db_session, address, raise_exc=True):
    """
    Find a server by name
    """
    query = db_session.query(ServerSM)\
            .filter(ServerSM.address == address)
    try:
        server_sm = query.one()
    except NoResultFound:
        server_sm = None
    if raise_exc and not server_sm:
        raise NoServerFoundByAddress(address)
    return server_sm

def rename_server(db_session, server_name=None, new_server_name=None,
        server_sm=None):
    """
    Rename a server
    """
    if not server_sm:
        server_sm = find_server_byname(db_session, server_name)
    new_server_name = new_server_name.lower()
    if new_server_name.endswith('.'):
        new_server_name = server_name[:-1]
    try:
        result = db_session.query(ServerSM)\
                .filter(ServerSM.name == new_server_name).all()
    except NoResultFound:
        pass
    if len(result):
        raise ServerExists(new_server_name)
    server_sm.name = new_server_name

def set_server_ssh_address(db_session, server_name, ssh_address):
    """
    Set a servers ssh_address
    """
    server_sm = find_server_byname(db_session, server_name)
    server_sm.ssh_address = ssh_address
    db_session.flush()

def set_server(db_session, server_name, new_server_name=None, address=None,
                server_type=None, ssh_address=None):
    """
    Change a servers data.  Can on only be done when it is disabled.
    """
    server_sm = find_server_byname(db_session, server_name)
    if not server_sm.is_disabled():
        raise ServerNotDisabled(server_sm.name)
    if address:
        server_sm.address = address
    if ssh_address:
        server_sm.ssh_address = ssh_address
    if new_server_name:
        rename_server(db_session, new_server_name=new_server_name,
                server_sm=server_sm)
    if server_type:
        server_sm.server_type = server_type
    db_session.flush()

def move_server_sg(db_session, server_name, sg_name=None):
    """
    Move a server between SGs
    """
    server_sm = find_server_byname(db_session, server_name)
    if not server_sm.is_disabled():
        raise ServerNotDisabled(server_sm.name)
    if not sg_name:
        sg_name = zone_cfg.get_row_exc(db_session, 'default_sg')
    if not sg_name in list_all_sgs(db_session):
        raise NoSgFound(sg_name)
    sg = find_sg_byname(db_session, sg_name, raise_exc=True)
    server_sm.set_sg(sg)

# Set up SOA query rcodes
_soaquery_rcodes = {}
def init_soaquery_rcodes():
    """
    Setup SOA query keep alive rcodes
    """
    #Transform settings for DYNDNS RCODES to something we can understand
    _soaquery_rcodes['success'] = [dns.rcode.from_text(x) 
        for x in 
            settings['serversm_soaquery_success_rcodes'].strip().split()]
    _soaquery_rcodes['ok'] = [dns.rcode.from_text(x) 
        for x in settings['serversm_soaquery_ok_rcodes'].strip().split()]
    _soaquery_rcodes['retry'] = [dns.rcode.from_text(x) 
        for x in settings['serversm_soaquery_retry_rcodes'].strip().split()]
    _soaquery_rcodes['broken'] = [dns.rcode.from_text(x) 
        for x in 
            settings['serversm_soaquery_broken_rcodes'].strip().split()]





#!/usr/bin/env python3.2
#
# Copyright (c) Net24 Limited, Christchurch, New Zealand 2011-2012
#       and     Voyager Internet Ltd, New Zealand, 2012-2013
#
#    This file is part of py-magcode-core.
#
#    Py-magcode-core is free software: you can redistribute it and/or modify
#    it under the terms of the GNU  General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Py-magcode-core is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU  General Public License for more details.
#
#    You should have received a copy of the GNU  General Public License
#    along with py-magcode-core.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Zone security tag class, corresponding to zone_sectags table
"""

from sqlalchemy.orm.exc import NoResultFound

from magcode.core.database import *
from dms.exceptions import ZoneSecTagExists
from dms.exceptions import ZoneSecTagDoesNotExist
from dms.exceptions import ZoneSecTagStillUsed

@saregister
class ZoneSecTag(object):
    """
    DNS Resource Record comment.

    Comparison methods are also used for sorting displayed output.
    """
    _table="zone_sectags"

    def __init__(self, sectag_label=None):
        """
        Initialise a security tag comment
        """
        self.sectag = sectag_label

    # For comparison purposes, including display!
    def __eq__(self, other):
        return self.sectag == other.sectag

    def __ne__(self, other):
        return self.sectag != other.sectag

    def __lt__(self, other):
        return self.sectag < other.sectag
    
    def __gt__(self, other):
        return self.sectag > other.sectag
    
    def __le__(self, other):
        return self.sectag <= other.sectag
    
    def __ge__(self, other):
        return self.sectag >= other.sectag

    def __str__(self):
        """
        Print out sectag name
        """
        return str(self.sectag)

    def to_engine(self, time_format=None):
        """
        Output for zone engine.
        """
        return {'zone_id': self.sectag, 'sectag_label': self.sectag}

    def to_engine_brief(self, time_format=None):
        """
        Brief output for zone_engine
        """
        return {'sectag_label': self.sectag}

def new_sectag(db_session, sectag_label):
    """
    Create a new sectag type
    """
    if sectag_label == settings['admin_sectag']:
        raise ZoneSecTagExists(sectag_label)
    zone_sectag = ZoneSecTag(sectag_label)
    try:
        sectag_list = db_session.query(ZoneSecTag)\
                        .filter(ZoneSecTag.zone_id == None)\
                        .filter(ZoneSecTag.sectag == sectag_label).all()
        if len(sectag_list):
            raise ZoneSecTagExists(sectag_label)
    except NoResultFound:
        pass
    db_session.add(zone_sectag)
    db_session.flush()
    return zone_sectag

def del_sectag(db_session, sectag_label):
    """
    Delete a sectag label
    """
    if sectag_label == settings['admin_sectag']:
        raise ZoneSecTagStillUsed(sectag_label)
    zone_sectag = ZoneSecTag(sectag_label)
    try:
        zone_sectag = db_session.query(ZoneSecTag)\
                        .filter(ZoneSecTag.zone_id == None)\
                        .filter(ZoneSecTag.sectag == sectag_label).one()
    except NoResultFound:
        raise ZoneSecTagDoesNotExist(sectag_label)
    # Check that it is no longer being used.
    try:
        in_use_count = db_session.query(ZoneSecTag.sectag)\
                        .filter(ZoneSecTag.zone_id != None)\
                        .filter(ZoneSecTag.sectag == sectag_label).count()
        if in_use_count:
            raise ZoneSecTagStillUsed(sectag_label)
    except NoResultFound:
        pass
    db_session.delete(zone_sectag)
    db_session.flush()
    del zone_sectag

def list_all_sectags(db_session):
    """
    Return list of all sectags
    """
    zone_sectags = [ZoneSecTag(settings['admin_sectag'])]
    try:
        zone_sectags.extend(db_session.query(ZoneSecTag)\
                            .filter(ZoneSecTag.zone_id == None).all())
    except NoResultFound:
        return zone_sectags
    return zone_sectags

def list_all_sectag_labels(db_session):
    """
    Return a list of all the sectag labels
    """
    zone_sectag_labels = [settings['admin_sectag']]
    try:
        zone_sectag_label_list = db_session.query(ZoneSecTag.sectag)\
                            .filter(ZoneSecTag.zone_id == None).all()
    except NoResultFound:
        pass
    zone_sectag_labels.extend([x[0] for x in zone_sectag_label_list])
    return zone_sectag_labels



#!/usr/bin/env python3.2
#
# Copyright (c) Net24 Limited, Christchurch, New Zealand 2011-2012
#       and     Voyager Internet Ltd, New Zealand, 2012-2013
#
#    This file is part of py-magcode-core.
#
#    Py-magcode-core is free software: you can redistribute it and/or modify
#    it under the terms of the GNU  General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Py-magcode-core is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU  General Public License for more details.
#
#    You should have received a copy of the GNU  General Public License
#    along with py-magcode-core.  If not, see <http://www.gnu.org/licenses/>.
#
"""Test Utility

Program implemented by subclassing magcode.core.process.Process, and 
replacing the main() method.
"""

import os
import sys

from magcode.core.process import Process
from magcode.core.globals_ import *
from magcode.core.database import *
from dms.globals_ import *
from dms.database.zone_instance import ZoneInstance
from dms.database.zone_sm import ZoneSM


settings['config_section'] = 'DEFAULT'

class DmsSaSandpitApp(Process):
    """
    Process Main Daemon class
    """
    
#    def parse_argv_left(self, argv_left):
#        """
#        Handle any arguments left after processing all switches
#
#        Override in application if needed.
#        """
#        if (len(argv_left) == 0):
#            self.usage_short()
#            sys.exit(os.EX_USAGE)
#        
#        self.argv_left = argv_left
#
    def main_process(self):
        """Main process for dms_test_sa_sandpit
        """
        # Connect to database, intialise SQL Alchemy
        setup_sqlalchemy()
        db_session = sql_data['scoped_session_class']()

        import pdb; pdb.set_trace()

        sys.exit(os.EX_OK)





if (__name__ is "__main__"):
    exit_code = DmsSaSandpitApp(sys.argv, len(sys.argv))
    sys.exit(exit_code)

